// selectionGroup.cpp

#include "CvGameCoreDLL.h"
#include "CvGlobals.h"
#include "CvSelectionGroup.h"
#include "CvGameAI.h"
#include "CvPlayerAI.h"
#include "CvTeamAI.h"
#include "CvUnit.h"
#include "CvGameCoreUtils.h"
#include "CvMap.h"
#include "CvPlot.h"
#include "CvDLLEntityIFaceBase.h"
#include "CvDLLInterfaceIFaceBase.h"
#include "CvDLLFAStarIFaceBase.h"
#include "FAStarNode.h"
#include "CvInfos.h"
#include "FProfiler.h"
#include "CvDLLEventReporterIFaceBase.h"

// Public Functions...

CvSelectionGroup::CvSelectionGroup()
{
	reset(0, NO_PLAYER, true);
}


CvSelectionGroup::~CvSelectionGroup()
{
	uninit();
}


void CvSelectionGroup::init(int iID, PlayerTypes eOwner)
{
	//--------------------------------
	// Init saved data
	reset(iID, eOwner);

	//--------------------------------
	// Init non-saved data

	//--------------------------------
	// Init other game data
	AI_init();
}


void CvSelectionGroup::uninit()
{
	m_units.clear();

	m_missionQueue.clear();
}


// FUNCTION: reset()
// Initializes data members that are serialized.
void CvSelectionGroup::reset(int iID, PlayerTypes eOwner, bool bConstructorCall)
{
	//--------------------------------
	// Uninit class
	uninit();

	m_iID = iID;
	m_iMissionTimer = 0;

	m_bForceUpdate = false;

	m_eOwner = eOwner;

	m_eActivityType = ACTIVITY_AWAKE;
	m_eAutomateType = NO_AUTOMATE;

	if (!bConstructorCall)
	{
		AI_reset();
	}
}


void CvSelectionGroup::kill()
{
	FAssert(getOwnerINLINE() != NO_PLAYER);
	FAssertMsg(getID() != FFreeList::INVALID_INDEX, "getID() is not expected to be equal with FFreeList::INVALID_INDEX");
	FAssertMsg(getNumUnits() == 0, "The number of units is expected to be 0");

	GET_PLAYER(getOwnerINLINE()).removeGroupCycle(getID());

	GET_PLAYER(getOwnerINLINE()).deleteSelectionGroup(getID());
}


void CvSelectionGroup::doTurn()
{
	PROFILE("CvSelectionGroup::doTurn()")

	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	bool bHurt;
	int iWaitTurns;
	int iBestWaitTurns;

	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (getNumUnits() > 0)
	{
		bHurt = false;

		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			pLoopUnit->doTurn();

			if (pLoopUnit->isHurt())
			{
				bHurt = true;
			}
		}

		if ((getActivityType() == ACTIVITY_HOLD) ||
			  ((getActivityType() == ACTIVITY_HEAL) && !bHurt) ||
			  ((getActivityType() == ACTIVITY_SENTRY) && (GET_PLAYER(getOwnerINLINE()).AI_getPlotDanger(plot(), 1) > 0)))
		{
			setActivityType(ACTIVITY_AWAKE);
		}

		if (AI_isControlled())
		{
			if ((getActivityType() != ACTIVITY_MISSION) || (!canFight() && (GET_PLAYER(getOwnerINLINE()).AI_getPlotDanger(plot()) > 0)))
			{
				setForceUpdate(true);
			}
		}
		else
		{
			if (getActivityType() == ACTIVITY_MISSION)
			{
				if (GET_PLAYER(getOwnerINLINE()).AI_getPlotDanger(plot(), 1) > 0)
				{
					clearMissionQueue();
				}
			}
		}

		if (isHuman())
		{
			if (GC.getGameINLINE().isMPOption(MPOPTION_SIMULTANEOUS_TURNS))
			{
				iBestWaitTurns = 0;

				pUnitNode = headUnitNode();

				while (pUnitNode != NULL)
				{
					pLoopUnit = ::getUnit(pUnitNode->m_data);
					pUnitNode = nextUnitNode(pUnitNode);

					iWaitTurns = (GC.getDefineINT("MIN_TIMER_UNIT_DOUBLE_MOVES") - (GC.getGameINLINE().getTurnSlice() - pLoopUnit->getLastMoveTurn()));

					if (iWaitTurns > iBestWaitTurns)
					{
						iBestWaitTurns = iWaitTurns;
					}
				}

				setMissionTimer(max(iBestWaitTurns, getMissionTimer()));
			}
		}
	}

	doDelayedDeath();
}


void CvSelectionGroup::updateTimers()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	bool bCombat;

	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (getNumUnits() > 0)
	{
		bCombat = false;

		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			if (pLoopUnit->isCombat())
			{
				// Dale - Stack Attack: use current update combat method
				if(DCMStackAttack)
				{
					if(getOwnerINLINE() == GC.getGameINLINE().getActivePlayer())
					{
                        pLoopUnit->updateStackCombat(pLoopUnit->getCombatUnit(), 2);
					}
				} else {
					pLoopUnit->updateCombat();
				}
				bCombat = true;
				break;
			}
		}

		if (!bCombat)
		{
			updateMission();
		}
	}

	doDelayedDeath();
}


// Returns true if group was killed...
bool CvSelectionGroup::doDelayedDeath()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (isBusy())
	{
		return false;
	}

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		pLoopUnit->doDelayedDeath();
	}

	if (getNumUnits() == 0)
	{
		kill();
		return true;
	}

	return false;
}


void CvSelectionGroup::playActionSound()
{
	// Pitboss should not be playing sounds!
#ifndef PITBOSS

	CvUnit *pHeadUnit;
	int iScriptId = -1;

	pHeadUnit = getHeadUnit();
	if ( pHeadUnit )
	{
		CvUnitInfo *pUnitInfo;
		pUnitInfo = &GC.getUnitInfo( pHeadUnit->getUnitType() );
		if ( pUnitInfo )
		{
			iScriptId = pUnitInfo->getArtInfo(0, GET_PLAYER(getOwner()).getCurrentEra())->getActionSoundScriptId();
		}
	}

	if ( (iScriptId == -1) && pHeadUnit )
	{
		CvCivilizationInfo *pCivInfo;
		pCivInfo = &GC.getCivilizationInfo( pHeadUnit->getCivilizationType() );
		if ( pCivInfo )
		{
			iScriptId = pCivInfo->getActionSoundScriptId();
		}
	}

	if ( (iScriptId != -1) && pHeadUnit )
	{
		CvPlot *pPlot = GC.getMapINLINE().plotINLINE(pHeadUnit->getX_INLINE(),pHeadUnit->getY_INLINE());
		if ( pPlot )
		{
			gDLL->Do3DSound( iScriptId, pPlot->getPoint() );
		}
	}

#endif // n PITBOSS
}


void CvSelectionGroup::pushMission(MissionTypes eMission, int iData1, int iData2, int iFlags, bool bAppend, bool bManual, MissionAITypes eMissionAI, CvPlot* pMissionAIPlot, CvUnit* pMissionAIUnit)
{
	PROFILE_FUNC();

	MissionData mission;

	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (!bAppend)
	{
		if (isBusy())
		{
			return;
		}

		clearMissionQueue();
	}

	if (bManual)
	{
		setAutomateType(NO_AUTOMATE);
	}

	mission.eMissionType = eMission;
	mission.iData1 = iData1;
	mission.iData2 = iData2;
	mission.iFlags = iFlags;
	mission.iPushTurn = GC.getGameINLINE().getGameTurn();

	AI_setMissionAI(eMissionAI, pMissionAIPlot, pMissionAIUnit);

	insertAtEndMissionQueue(mission, !bAppend);

	if (bManual)
	{
		if (getOwnerINLINE() == GC.getGameINLINE().getActivePlayer())
		{
			if (isBusy() && GC.getMissionInfo(eMission).isSound())
			{
				playActionSound();
			}

			gDLL->getInterfaceIFace()->setHasMovedUnit(true);
		}

		gDLL->getEventReporterIFace()->selectionGroupPushMission(this, eMission);

		doDelayedDeath();
	}
}


void CvSelectionGroup::popMission()
{
	CLLNode<MissionData>* pTailNode;

	FAssert(getOwnerINLINE() != NO_PLAYER);

	pTailNode = tailMissionQueueNode();

	if (pTailNode != NULL)
	{
		deleteMissionQueueNode(pTailNode);
	}
}


void CvSelectionGroup::autoMission()
{
	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (getNumUnits() > 0)
	{
		if (headMissionQueueNode() != NULL)
		{
			if (!isBusy())
			{
				if (getActivityType() == ACTIVITY_MISSION)
				{
					continueMission();
				}
				else
				{
					startMission();
				}
			}
		}
	}

	doDelayedDeath();
}


void CvSelectionGroup::updateMission()
{
	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (getMissionTimer() > 0)
	{
		changeMissionTimer(-1);

		if (getMissionTimer() == 0)
		{
			if (getActivityType() == ACTIVITY_MISSION)
			{
				continueMission();
			}
			else
			{
				if (getOwnerINLINE() == GC.getGameINLINE().getActivePlayer())
				{
					if (gDLL->getInterfaceIFace()->getHeadSelectedUnit() == NULL)
					{
						gDLL->getInterfaceIFace()->changeCycleSelectionCounter(1);
					}
				}
			}
		}
	}
}


CvPlot* CvSelectionGroup::lastMissionPlot()
{
	CLLNode<MissionData>* pMissionNode;
	CvUnit* pTargetUnit;

	pMissionNode = tailMissionQueueNode();

	while (pMissionNode != NULL)
	{
		switch (pMissionNode->m_data.eMissionType)
		{
		case MISSION_MOVE_TO:
		case MISSION_ROUTE_TO:
			return GC.getMapINLINE().plotINLINE(pMissionNode->m_data.iData1, pMissionNode->m_data.iData2);
			break;

		case MISSION_MOVE_TO_UNIT:
			pTargetUnit = GET_PLAYER((PlayerTypes)pMissionNode->m_data.iData1).getUnit(pMissionNode->m_data.iData2);
			if (pTargetUnit != NULL)
			{
				return pTargetUnit->plot();
			}
			break;

		case MISSION_SKIP:
		case MISSION_SLEEP:
		case MISSION_FORTIFY:
		case MISSION_AIRPATROL:
		case MISSION_HEAL:
		case MISSION_SENTRY:
		case MISSION_AIRLIFT:
		case MISSION_NUKE:
		case MISSION_RECON:
		case MISSION_AIRBOMB:
		case MISSION_BOMBARD:
		case MISSION_PILLAGE:

		// < Route Pillage Start >
		case MISSION_PILLAGE_ROUTE:
		// < Route Pillage End   >

		case MISSION_SABOTAGE:
		case MISSION_DESTROY:
		case MISSION_STEAL_PLANS:
		case MISSION_FOUND:
		case MISSION_SPREAD:
		case MISSION_JOIN:
		case MISSION_CONSTRUCT:
		case MISSION_DISCOVER:
		case MISSION_HURRY:
		case MISSION_TRADE:
		case MISSION_GREAT_WORK:
		case MISSION_GOLDEN_AGE:
		case MISSION_BUILD:
		case MISSION_LEAD:
			break;

		default:
			FAssert(false);
			break;
		}

		pMissionNode = prevMissionQueueNode(pMissionNode);
	}

	return plot();
}


bool CvSelectionGroup::canStartMission(int iMission, int iData1, int iData2, CvPlot* pPlot, bool bTestVisible)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pTargetUnit;
	CvUnit* pLoopUnit;

	if (isBusy())
	{
		return false;
	}

	if (pPlot == NULL)
	{
		pPlot = plot();
	}

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		switch (iMission)
		{
		case MISSION_MOVE_TO:
			if (!(pPlot->at(iData1, iData2)))
			{
				return true;
			}
			break;

		case MISSION_ROUTE_TO:
			if (!(pPlot->at(iData1, iData2)) || (getBestBuildRoute(pPlot) != NO_ROUTE))
			{
				return true;
			}
			break;

		case MISSION_MOVE_TO_UNIT:
			FAssertMsg(iData1 != NO_PLAYER, "iData1 should be a valid Player");
			pTargetUnit = GET_PLAYER((PlayerTypes)iData1).getUnit(iData2);
			if ((pTargetUnit != NULL) && !(pTargetUnit->atPlot(pPlot)))
			{
				return true;
			}
			break;

		case MISSION_SKIP:
			if (pLoopUnit->canHold(pPlot))
			{
				return true;
			}
			break;

		case MISSION_SLEEP:
			if (pLoopUnit->canSleep(pPlot))
			{
				return true;
			}
			break;

		case MISSION_FORTIFY:
			if (pLoopUnit->canFortify(pPlot))
			{
				return true;
			}
			break;

		case MISSION_AIRPATROL:
			if (pLoopUnit->canAirPatrol(pPlot))
			{
				return true;
			}
			break;

		case MISSION_HEAL:
			if (pLoopUnit->canHeal(pPlot))
			{
				return true;
			}
			break;

		case MISSION_SENTRY:
			if (pLoopUnit->canSentry(pPlot))
			{
				return true;
			}
			break;

		case MISSION_AIRLIFT:
			if (pLoopUnit->canAirliftAt(pPlot, iData1, iData2))
			{
				return true;
			}
			break;

		case MISSION_NUKE:
			if (pLoopUnit->canNukeAt(pPlot, iData1, iData2))
			{
				return true;
			}
			break;

		case MISSION_RECON:
			if (pLoopUnit->canReconAt(pPlot, iData1, iData2))
			{
				return true;
			}
			break;

		case MISSION_AIRBOMB:
			if (pLoopUnit->canAirBombAt(pPlot, iData1, iData2))
			{
				return true;
			}
			break;

		case MISSION_BOMBARD:
			if(DCMFieldBombard)	// Dale - Field Bombard:
			{
				if (pLoopUnit->canBombardAtRanged(pPlot, iData1, iData2))	// Dale - Field Bombard: changed for ranged bombard
				{
					return true;
				}
			} else {
				if (pLoopUnit->canBombard(pPlot))
				{
					return true;
				}
			}
			break;

		case MISSION_PILLAGE:
			if (pLoopUnit->canPillage(pPlot))
			{
				return true;
			}
			break;

		// < Route Pillage Start >
		case MISSION_PILLAGE_ROUTE:
			if (pLoopUnit->canPillageRoute(pPlot))
			{
				return true;
			}
			break;
		// < Route Pillage End   >

		case MISSION_SABOTAGE:
			if (pLoopUnit->canSabotage(pPlot, bTestVisible))
			{
				return true;
			}
			break;

		case MISSION_DESTROY:
			if (pLoopUnit->canDestroy(pPlot, bTestVisible))
			{
				return true;
			}
			break;

		case MISSION_STEAL_PLANS:
			if (pLoopUnit->canStealPlans(pPlot, bTestVisible))
			{
				return true;
			}
			break;

		case MISSION_FOUND:
			if (pLoopUnit->canFound(pPlot, bTestVisible))
			{
				return true;
			}
			break;

		case MISSION_SPREAD:
			if (pLoopUnit->canSpread(pPlot, ((ReligionTypes)iData1), bTestVisible))
			{
				return true;
			}
			break;

		case MISSION_JOIN:
			if (pLoopUnit->canJoin(pPlot, ((SpecialistTypes)iData1)))
			{
				return true;
			}
			break;

		case MISSION_CONSTRUCT:
			if (pLoopUnit->canConstruct(pPlot, ((BuildingTypes)iData1), bTestVisible))
			{
				return true;
			}
			break;

		case MISSION_DISCOVER:
			if (pLoopUnit->canDiscover(pPlot))
			{
				return true;
			}
			break;

		case MISSION_HURRY:
			if (pLoopUnit->canHurry(pPlot, bTestVisible))
			{
				return true;
			}
			break;

		case MISSION_TRADE:
			if (pLoopUnit->canTrade(pPlot, bTestVisible))
			{
				return true;
			}
			break;

		case MISSION_GREAT_WORK:
			if (pLoopUnit->canGreatWork(pPlot))
			{
				return true;
			}
			break;

		case MISSION_GOLDEN_AGE:
			if (pLoopUnit->canGoldenAge(pPlot, bTestVisible))
			{
				return true;
			}
			break;

		case MISSION_BUILD:
			if (pLoopUnit->canBuild(pPlot, ((BuildTypes)iData1), bTestVisible))
			{
				return true;
			}
			break;

		case MISSION_LEAD:
			if (pLoopUnit->canLead(pPlot, iData1))
			{
				return true;
			}
			break;

		case MISSION_BEGIN_COMBAT:
		case MISSION_END_COMBAT:
		case MISSION_AIRSTRIKE:
		case MISSION_SURRENDER:
		case MISSION_IDLE:
		case MISSION_DIE:
		case MISSION_DAMAGE:
		case MISSION_MULTI_SELECT:
		case MISSION_MULTI_DESELECT:
			break;

		default:
			FAssert(false);
			break;
		}
	}

	return false;
}


void CvSelectionGroup::startMission()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	bool bDelete;
	bool bAction;
	bool bNuke;
	bool bNotify;

	FAssert(!isBusy());
	FAssert(getOwnerINLINE() != NO_PLAYER);
	FAssert(headMissionQueueNode() != NULL);

	if (!(GC.getGameINLINE().isMPOption(MPOPTION_SIMULTANEOUS_TURNS)))
	{
		if (!(GET_PLAYER(getOwnerINLINE()).isTurnActive()))
		{
			if (getOwnerINLINE() == GC.getGameINLINE().getActivePlayer())
			{
				if (IsSelected())
				{
					gDLL->getInterfaceIFace()->changeCycleSelectionCounter(1);
				}
			}

			return;
		}
	}

	setActivityType(ACTIVITY_MISSION);

	bDelete = false;
	bAction = false;
	bNuke = false;
	bNotify = false;

	if (!canStartMission(headMissionQueueNode()->m_data.eMissionType, headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2, plot()))
	{
		bDelete = true;
	}
	else
	{
		FAssertMsg(GET_PLAYER(getOwnerINLINE()).isTurnActive() || GET_PLAYER(getOwnerINLINE()).isHuman(), "It's expected that either the turn is active for this player or the player is human");

		switch (headMissionQueueNode()->m_data.eMissionType)
		{
		case MISSION_MOVE_TO:
		case MISSION_ROUTE_TO:
		case MISSION_MOVE_TO_UNIT:
			break;

		case MISSION_SKIP:
			setActivityType(ACTIVITY_HOLD);
			bDelete = true;
			break;

		case MISSION_SLEEP:
			setActivityType(ACTIVITY_SLEEP);
			bNotify = true;
			bDelete = true;
			break;

		case MISSION_FORTIFY:
			setActivityType(ACTIVITY_SLEEP);
			bNotify = true;
			bDelete = true;
			break;

		case MISSION_AIRPATROL:
			setActivityType(ACTIVITY_INTERCEPT);
			bDelete = true;
			break;

		case MISSION_HEAL:
			setActivityType(ACTIVITY_HEAL);
			bNotify = true;
			bDelete = true;
			break;

		case MISSION_SENTRY:
			setActivityType(ACTIVITY_SENTRY);
			bNotify = true;
			bDelete = true;
			break;

		case MISSION_AIRLIFT:
		case MISSION_NUKE:
		case MISSION_RECON:
		case MISSION_AIRBOMB:
		case MISSION_BOMBARD:
		case MISSION_PILLAGE:

		// < Route Pillage Start >
		case MISSION_PILLAGE_ROUTE:
		// < Route Pillage End   >

		case MISSION_SABOTAGE:
		case MISSION_DESTROY:
		case MISSION_STEAL_PLANS:
		case MISSION_FOUND:
		case MISSION_SPREAD:
		case MISSION_JOIN:
		case MISSION_CONSTRUCT:
		case MISSION_DISCOVER:
		case MISSION_HURRY:
		case MISSION_TRADE:
		case MISSION_GREAT_WORK:
		case MISSION_GOLDEN_AGE:
		case MISSION_BUILD:
		case MISSION_LEAD:
			break;

		default:
			FAssert(false);
			break;
		}

		if ( bNotify )
		{
			NotifyEntity( headMissionQueueNode()->m_data.eMissionType );
		}

		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			if (pLoopUnit->canMove())
			{
				switch (headMissionQueueNode()->m_data.eMissionType)
				{
				case MISSION_MOVE_TO:
				case MISSION_ROUTE_TO:
				case MISSION_MOVE_TO_UNIT:
				case MISSION_SKIP:
				case MISSION_SLEEP:
				case MISSION_FORTIFY:
				case MISSION_AIRPATROL:
				case MISSION_HEAL:
				case MISSION_SENTRY:
					break;

				case MISSION_AIRLIFT:
					if (pLoopUnit->airlift(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2))
					{
						bAction = true;
					}
					break;

				case MISSION_NUKE:
					if (pLoopUnit->nuke(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2))
					{
						bAction = true;

						if (GC.getMapINLINE().plotINLINE(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2)->isVisibleToWatchingHuman())
						{
							bNuke = true;
						}
					}
					break;

				case MISSION_RECON:
					if (pLoopUnit->recon(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2))
					{
						bAction = true;
					}
					break;

				case MISSION_AIRBOMB:
					if (pLoopUnit->airBomb(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2))
					{
						bAction = true;
					}
					break;

				case MISSION_BOMBARD:
					if(DCMFieldBombard)				// Dale - Field Bombard: new method
					{
						if (pLoopUnit->bombardRanged(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2))
						{
							bAction = true;
						}
					} else {
						if(pLoopUnit->bombard())	// Dale - Field Bombard: old method
						{
							bAction = true;
						}
					}
					break;

				case MISSION_PILLAGE:
					if (pLoopUnit->pillage())
					{
						bAction = true;
					}
					break;

				// < Route Pillage Start >

				case MISSION_PILLAGE_ROUTE:
					if (pLoopUnit->pillageRoute())
					{
						bAction = true;
					}
					break;
				// < Route Pillage End   >

				case MISSION_SABOTAGE:
					if (pLoopUnit->sabotage())
					{
						bAction = true;
					}
					break;

				case MISSION_DESTROY:
					if (pLoopUnit->destroy())
					{
						bAction = true;
					}
					break;

				case MISSION_STEAL_PLANS:
					if (pLoopUnit->stealPlans())
					{
						bAction = true;
					}
					break;

				case MISSION_FOUND:
					if (pLoopUnit->found())
					{
						bAction = true;
					}
					break;

				case MISSION_SPREAD:
					if (pLoopUnit->spread((ReligionTypes)(headMissionQueueNode()->m_data.iData1)))
					{
						bAction = true;
					}
					break;

				case MISSION_JOIN:
					if (pLoopUnit->join((SpecialistTypes)(headMissionQueueNode()->m_data.iData1)))
					{
						bAction = true;
					}
					break;

				case MISSION_CONSTRUCT:
					if (pLoopUnit->construct((BuildingTypes)(headMissionQueueNode()->m_data.iData1)))
					{
						bAction = true;
					}
					break;

				case MISSION_DISCOVER:
					if (pLoopUnit->discover())
					{
						bAction = true;
					}
					break;

				case MISSION_HURRY:
					if (pLoopUnit->hurry())
					{
						bAction = true;
					}
					break;

				case MISSION_TRADE:
					if (pLoopUnit->trade())
					{
						bAction = true;
					}
					break;

				case MISSION_GREAT_WORK:
					if (pLoopUnit->greatWork())
					{
						bAction = true;
					}
					break;

				case MISSION_GOLDEN_AGE:
					if (pLoopUnit->goldenAge())
					{
						bAction = true;
					}
					break;

				case MISSION_BUILD:
					break;

				case MISSION_LEAD:
					if (pLoopUnit->lead(headMissionQueueNode()->m_data.iData1))
					{
						bAction = true;
					}
					break;

				default:
					FAssert(false);
					break;
				}

				if (getNumUnits() == 0)
				{
					break;
				}

				if (headMissionQueueNode() == NULL)
				{
					break;
				}
			}
		}
	}

	if ((getNumUnits() > 0) && (headMissionQueueNode() != NULL))
	{
		if (bAction)
		{
			if (isHuman())
			{
				if (plot()->isVisibleToWatchingHuman())
				{
					updateMissionTimer();
				}
			}
		}

		if (bNuke)
		{
			setMissionTimer(GC.getMissionInfo(MISSION_NUKE).getTime());
		}

		if (!isBusy())
		{
			if (bDelete)
			{
				if (getOwnerINLINE() == GC.getGameINLINE().getActivePlayer())
				{
					if (IsSelected())
					{
						gDLL->getInterfaceIFace()->changeCycleSelectionCounter((GET_PLAYER(getOwnerINLINE()).isOption(PLAYEROPTION_QUICK_MOVES)) ? 1 : 2);
					}
				}

				deleteMissionQueueNode(headMissionQueueNode());
			}
			else if (getActivityType() == ACTIVITY_MISSION)
			{
				continueMission();
			}
		}
	}
}


void CvSelectionGroup::continueMission(int iSteps)
{
	CvUnit* pTargetUnit;
	bool bDone;
	bool bAction;

	FAssert(!isBusy());
	FAssert(headMissionQueueNode() != NULL);
	FAssert(getOwnerINLINE() != NO_PLAYER);
	FAssert(getActivityType() == ACTIVITY_MISSION);

	if (headMissionQueueNode() == NULL)
	{
		// just in case...
		setActivityType(ACTIVITY_AWAKE);
		return;
	}

	bDone = false;
	bAction = false;

	if (headMissionQueueNode()->m_data.iPushTurn == GC.getGameINLINE().getGameTurn())
	{
		if (headMissionQueueNode()->m_data.eMissionType == MISSION_MOVE_TO)
		{
			// Dale - Stack Attack: use required group attack method whether concept on or off
			if(DCMStackAttack)
			{
				if (groupStackAttack(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2, headMissionQueueNode()->m_data.iFlags))
				{
					bDone = true;
				}
			} else {
				if (groupAttack(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2, headMissionQueueNode()->m_data.iFlags))
				{
					bDone = true;
				}
			}
		}
	}

	if (canAllMove())
	{
		if (getNumUnits() > 0)
		{
			if (!bDone && canAllMove())
			{
				switch (headMissionQueueNode()->m_data.eMissionType)
				{
				case MISSION_MOVE_TO:
					if (getDomainType() == DOMAIN_AIR)
					{
						groupPathTo(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2, headMissionQueueNode()->m_data.iFlags);
						bDone = true;
					}
					else if (groupPathTo(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2, headMissionQueueNode()->m_data.iFlags))
					{
						bAction = true;

						if (getNumUnits() > 0)
						{
							if (!canAllMove())
							{
								if (headMissionQueueNode() != NULL)
								{
									if (groupAmphibMove(GC.getMapINLINE().plotINLINE(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2), headMissionQueueNode()->m_data.iFlags))
									{
										bAction = false;
										bDone = true;
									}
								}
							}
						}
					}
					else
					{
						bDone = true;
					}
					break;

				case MISSION_ROUTE_TO:
					if (groupRoadTo(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2, headMissionQueueNode()->m_data.iFlags))
					{
						bAction = true;
					}
					else
					{
						bDone = true;
					}
					break;

				case MISSION_MOVE_TO_UNIT:
					pTargetUnit = GET_PLAYER((PlayerTypes)headMissionQueueNode()->m_data.iData1).getUnit(headMissionQueueNode()->m_data.iData2);
					if ((pTargetUnit != NULL) && groupPathTo(pTargetUnit->getX_INLINE(), pTargetUnit->getY_INLINE(), headMissionQueueNode()->m_data.iFlags))
					{
						bAction = true;
					}
					else
					{
						bDone = true;
					}
					break;

				case MISSION_SKIP:
				case MISSION_SLEEP:
				case MISSION_FORTIFY:
				case MISSION_AIRPATROL:
				case MISSION_HEAL:
				case MISSION_SENTRY:
					FAssert(false);
					break;

				case MISSION_AIRLIFT:
				case MISSION_NUKE:
				case MISSION_RECON:
				case MISSION_AIRBOMB:
				case MISSION_BOMBARD:
				case MISSION_PILLAGE:

				// < Route Pillage Start >
				case MISSION_PILLAGE_ROUTE:
				// < Route Pillage End   >

				case MISSION_SABOTAGE:
				case MISSION_DESTROY:
				case MISSION_STEAL_PLANS:
				case MISSION_FOUND:
				case MISSION_SPREAD:
				case MISSION_JOIN:
				case MISSION_CONSTRUCT:
				case MISSION_DISCOVER:
				case MISSION_HURRY:
				case MISSION_TRADE:
				case MISSION_GREAT_WORK:
				case MISSION_GOLDEN_AGE:
				case MISSION_LEAD:
					break;

				case MISSION_BUILD:
					if (!groupBuild((BuildTypes)(headMissionQueueNode()->m_data.iData1)))
					{
						bDone = true;
					}
					break;

				default:
					FAssert(false);
					break;
				}
			}
		}
	}

	if ((getNumUnits() > 0) && (headMissionQueueNode() != NULL))
	{
		if (!bDone)
		{
			switch (headMissionQueueNode()->m_data.eMissionType)
			{
			case MISSION_MOVE_TO:
				if (at(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2))
				{
					bDone = true;
				}
				break;

			case MISSION_ROUTE_TO:
				if (at(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2))
				{
					if (getBestBuildRoute(plot()) == NO_ROUTE)
					{
						bDone = true;
					}
				}
				break;

			case MISSION_MOVE_TO_UNIT:
				pTargetUnit = GET_PLAYER((PlayerTypes)headMissionQueueNode()->m_data.iData1).getUnit(headMissionQueueNode()->m_data.iData2);
				if ((pTargetUnit == NULL) || atPlot(pTargetUnit->plot()))
				{
					bDone = true;
				}
				break;

			case MISSION_SKIP:
			case MISSION_SLEEP:
			case MISSION_FORTIFY:
			case MISSION_AIRPATROL:
			case MISSION_HEAL:
			case MISSION_SENTRY:
				FAssert(false);
				break;

			case MISSION_AIRLIFT:
			case MISSION_NUKE:
			case MISSION_RECON:
			case MISSION_AIRBOMB:
			case MISSION_BOMBARD:
			case MISSION_PILLAGE:

			// < Route Pillage Start >
			case MISSION_PILLAGE_ROUTE:
			// < Route Pillage End   >

			case MISSION_SABOTAGE:
			case MISSION_DESTROY:
			case MISSION_STEAL_PLANS:
			case MISSION_FOUND:
			case MISSION_SPREAD:
			case MISSION_JOIN:
			case MISSION_CONSTRUCT:
			case MISSION_DISCOVER:
			case MISSION_HURRY:
			case MISSION_TRADE:
			case MISSION_GREAT_WORK:
			case MISSION_GOLDEN_AGE:
			case MISSION_LEAD:
				bDone = true;
				break;

			case MISSION_BUILD:
				// XXX what happens if two separate worker groups are both building the mine...
				/*if (plot()->getBuildType() != ((BuildTypes)(headMissionQueueNode()->m_data.iData1)))
				{
					bDone = true;
				}*/
				break;

			default:
				FAssert(false);
				break;
			}
		}
	}

	if ((getNumUnits() > 0) && (headMissionQueueNode() != NULL))
	{
		if (bAction)
		{
			if (bDone || !canAllMove())
			{
				if (plot()->isVisibleToWatchingHuman())
				{
					updateMissionTimer(iSteps);

					getHeadUnit()->ExecuteMove(((float)(getMissionTimer() * gDLL->getMillisecsPerTurn())) / 1000.0f);

					if (GC.getGameINLINE().showAllMoves(getTeam()))
					{
						if (GC.getGameINLINE().getActivePlayer() != NO_PLAYER)
						{
							if (getOwnerINLINE() != GC.getGameINLINE().getActivePlayer())
							{
								if (plot()->isActiveVisible(false))
								{
									gDLL->getInterfaceIFace()->lookAt(plot()->getPoint(), CAMERALOOKAT_NORMAL);
								}
							}
						}
					}
				}
			}
		}

		if (bDone)
		{
			if (!isBusy())
			{
				if (getOwnerINLINE() == GC.getGameINLINE().getActivePlayer())
				{
					if (IsSelected())
					{
						if ((headMissionQueueNode()->m_data.eMissionType == MISSION_MOVE_TO) ||
							  (headMissionQueueNode()->m_data.eMissionType == MISSION_ROUTE_TO) ||
							  (headMissionQueueNode()->m_data.eMissionType == MISSION_MOVE_TO_UNIT))
						{
							gDLL->getInterfaceIFace()->changeCycleSelectionCounter((GET_PLAYER(getOwnerINLINE()).isOption(PLAYEROPTION_QUICK_MOVES)) ? 1 : 2);
						}
					}
				}

				deleteMissionQueueNode(headMissionQueueNode());
			}
		}
		else
		{
			if (canAllMove())
			{
				continueMission(iSteps + 1);
			}
			else if (!isBusy())
			{
				if (getOwnerINLINE() == GC.getGameINLINE().getActivePlayer())
				{
					if (IsSelected())
					{
						gDLL->getInterfaceIFace()->changeCycleSelectionCounter(1);
					}
				}
			}
		}
	}
}


bool CvSelectionGroup::canDoCommand(CommandTypes eCommand, int iData1, int iData2, bool bTestVisible)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if (pLoopUnit->canDoCommand(eCommand, iData1, iData2, bTestVisible))
		{
			return true;
		}
	}

	return false;
}


// Returns true if one of the units can support the interface mode...
bool CvSelectionGroup::canDoInterfaceMode(InterfaceModeTypes eInterfaceMode)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	FAssertMsg(eInterfaceMode != NO_INTERFACEMODE, "InterfaceMode is not assigned a valid value");

	if (isBusy())
	{
		return false;
	}

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);

		switch (eInterfaceMode)
		{
		case INTERFACEMODE_GO_TO:
			if ((getDomainType() != DOMAIN_AIR) && (getDomainType() != DOMAIN_IMMOBILE))
			{
				return true;
			}
			break;

		case INTERFACEMODE_GO_TO_TYPE:
			if ((getDomainType() != DOMAIN_AIR) && (getDomainType() != DOMAIN_IMMOBILE))
			{
				if (pLoopUnit->plot()->plotCount(PUF_isUnitType, pLoopUnit->getUnitType(), -1, pLoopUnit->getOwnerINLINE()) > 1)
				{
					return true;
				}
			}
			break;

		case INTERFACEMODE_GO_TO_ALL:
			if ((getDomainType() != DOMAIN_AIR) && (getDomainType() != DOMAIN_IMMOBILE))
			{
				if (pLoopUnit->plot()->plotCount(NULL, -1, -1, pLoopUnit->getOwnerINLINE()) > 1)
				{
					return true;
				}
			}
			break;

		case INTERFACEMODE_ROUTE_TO:
			if (pLoopUnit->AI_getUnitAIType() == UNITAI_WORKER)
			{
				if (pLoopUnit->canBuildRoute())
				{
					return true;
				}
			}
			break;

		case INTERFACEMODE_AIRLIFT:
			if (pLoopUnit->canAirlift(pLoopUnit->plot()))
			{
				return true;
			}
			break;

		case INTERFACEMODE_NUKE:
			if (pLoopUnit->canNuke(pLoopUnit->plot()))
			{
				return true;
			}
			break;

		case INTERFACEMODE_RECON:
			if (pLoopUnit->canRecon(pLoopUnit->plot()))
			{
				return true;
			}
			break;

		case INTERFACEMODE_AIRBOMB:
			if (pLoopUnit->canAirBomb(pLoopUnit->plot()))
			{
				return true;
			}
			break;

		case INTERFACEMODE_AIRSTRIKE:
			if (pLoopUnit->getDomainType() == DOMAIN_AIR)
			{
				if (pLoopUnit->canAirAttack())
				{
					return true;
				}
			}
			break;

		case INTERFACEMODE_REBASE:
			if (pLoopUnit->getDomainType() == DOMAIN_AIR)
			{
				return true;
			}
			break;

		// Dale - Field Bombard: interface mode START
		case INTERFACEMODE_BOMBARD:
			if (pLoopUnit->canBombard(pLoopUnit->plot()) && DCMFieldBombard)
			{
				return true;
			}
			break;
		// Dale - Field Bombard: interface mode END

		}

		pUnitNode = nextUnitNode(pUnitNode);
	}

	return false;
}


// Returns true if one of the units can execute the interface mode at the specified plot...
bool CvSelectionGroup::canDoInterfaceModeAt(InterfaceModeTypes eInterfaceMode, CvPlot* pPlot)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	FAssertMsg(eInterfaceMode != NO_INTERFACEMODE, "InterfaceMode is not assigned a valid value");

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);

		switch (eInterfaceMode)
		{
		case INTERFACEMODE_AIRLIFT:
			if (pLoopUnit != NULL)
			{
				if (pLoopUnit->canAirliftAt(pLoopUnit->plot(), pPlot->getX_INLINE(), pPlot->getY_INLINE()))
				{
					return true;
				}
			}
			break;

		case INTERFACEMODE_NUKE:
			if (pLoopUnit != NULL)
			{
				if (pLoopUnit->canNukeAt(pLoopUnit->plot(), pPlot->getX_INLINE(), pPlot->getY_INLINE()))
				{
					return true;
				}
			}
			break;

		case INTERFACEMODE_RECON:
			if (pLoopUnit != NULL)
			{
				if (pLoopUnit->canReconAt(pLoopUnit->plot(), pPlot->getX_INLINE(), pPlot->getY_INLINE()))
				{
					return true;
				}
			}
			break;

		case INTERFACEMODE_AIRBOMB:
			if (pLoopUnit != NULL)
			{
				if (pLoopUnit->canAirBombAt(pLoopUnit->plot(), pPlot->getX_INLINE(), pPlot->getY_INLINE()))
				{
					return true;
				}
			}
			break;

		case INTERFACEMODE_AIRSTRIKE:
			if (pLoopUnit != NULL)
			{
				if (pLoopUnit->canMoveInto(pPlot, true))
				{
					return true;
				}
			}
			break;

		case INTERFACEMODE_REBASE:
			if (pLoopUnit != NULL)
			{
				if (pLoopUnit->canMoveInto(pPlot))
				{
					return true;
				}
			}
			break;

		// Dale - Field Bombard: interface mode START
		case INTERFACEMODE_BOMBARD:
			if (pLoopUnit != NULL)
			{
				if(DCMFieldBombard)
				{
					if (pLoopUnit->canBombardAtRanged(pLoopUnit->plot(), pPlot->getX_INLINE(), pPlot->getY_INLINE()))
					{
						return true;
					}
				}
			}
			break;
		// Dale - Field Bombard: interface mode END

		default:
			return true;
			break;
		}

		pUnitNode = nextUnitNode(pUnitNode);
	}

	return false;
}


bool CvSelectionGroup::isHuman()
{
	if (getOwnerINLINE() != NO_PLAYER)
	{
		return GET_PLAYER(getOwnerINLINE()).isHuman();
	}

	return true;
}


bool CvSelectionGroup::isBusy()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	CvPlot* pPlot;

	if (getNumUnits() == 0)
	{
		return false;
	}

	if (getMissionTimer() > 0)
	{
		return true;
	}

	pPlot = plot();

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if (pLoopUnit != NULL)
		{
			if (pLoopUnit->isCombat())
			{
				return true;
			}
		}
	}

	return false;
}


bool CvSelectionGroup::isCargoBusy()
{
	CLLNode<IDInfo>* pUnitNode1;
	CLLNode<IDInfo>* pUnitNode2;
	CvUnit* pLoopUnit1;
	CvUnit* pLoopUnit2;
	CvPlot* pPlot;

	if (getNumUnits() == 0)
	{
		return false;
	}

	pPlot = plot();

	pUnitNode1 = headUnitNode();

	while (pUnitNode1 != NULL)
	{
		pLoopUnit1 = ::getUnit(pUnitNode1->m_data);
		pUnitNode1 = nextUnitNode(pUnitNode1);

		if (pLoopUnit1 != NULL)
		{
			if (pLoopUnit1->getCargo() > 0)
			{
				pUnitNode2 = pPlot->headUnitNode();

				while (pUnitNode2 != NULL)
				{
					pLoopUnit2 = ::getUnit(pUnitNode2->m_data);
					pUnitNode2 = pPlot->nextUnitNode(pUnitNode2);

					if (pLoopUnit2->getTransportUnit() == pLoopUnit1)
					{
						if (pLoopUnit2->getGroup()->isBusy())
						{
							return true;
						}
					}
				}
			}
		}
	}

	return false;
}


int CvSelectionGroup::baseMoves()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	int iValue;
	int iBestValue;

	iBestValue = MAX_INT;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		iValue = pLoopUnit->baseMoves();

		if (iValue < iBestValue)
		{
			iBestValue = iValue;
		}
	}

	return iBestValue;
}


bool CvSelectionGroup::isWaiting() const
{
	return ((getActivityType() == ACTIVITY_HOLD) ||
		      (getActivityType() == ACTIVITY_SLEEP) ||
					(getActivityType() == ACTIVITY_HEAL) ||
					(getActivityType() == ACTIVITY_SENTRY) ||
					(getActivityType() == ACTIVITY_INTERCEPT));
}


bool CvSelectionGroup::isFull()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	if (getNumUnits() > 0)
	{
		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			if (!(pLoopUnit->isFull()))
			{
				return false;
			}
		}

		return true;
	}

	return false;
}


bool CvSelectionGroup::hasCargo()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if (pLoopUnit->hasCargo())
		{
			return true;
		}
	}

	return false;
}


bool CvSelectionGroup::canAllMove()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	if (getNumUnits() > 0)
	{
		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			if (!(pLoopUnit->canMove()))
			{
				return false;
			}
		}

		return true;
	}

	return false;
}


bool CvSelectionGroup::canAnyMove()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if (pLoopUnit->canMove())
		{
			return true;
		}
	}

	return false;
}


bool CvSelectionGroup::hasMoved()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if (pLoopUnit->hasMoved())
		{
			return true;
		}
	}

	return false;
}


bool CvSelectionGroup::canEnterTerritory(TeamTypes eTeam, bool bIgnoreRightOfPassage) const
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	if (getNumUnits() > 0)
	{
		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			if (!(pLoopUnit->canEnterTerritory(eTeam, bIgnoreRightOfPassage)))
			{
				return false;
			}
		}

		return true;
	}

	return false;
}

bool CvSelectionGroup::canEnterArea(TeamTypes eTeam, const CvArea* pArea, bool bIgnoreRightOfPassage) const
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	if (getNumUnits() > 0)
	{
		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			if (!(pLoopUnit->canEnterArea(eTeam, pArea, bIgnoreRightOfPassage)))
			{
				return false;
			}
		}

		return true;
	}

	return false;
}


bool CvSelectionGroup::canMoveInto(CvPlot* pPlot, bool bAttack)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	if (getNumUnits() > 0)
	{
		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			if (pLoopUnit->canMoveInto(pPlot, bAttack))
			{
				return true;
			}
		}
	}

	return false;
}


bool CvSelectionGroup::canMoveOrAttackInto(CvPlot* pPlot, bool bDeclareWar)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	if (getNumUnits() > 0)
	{
		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			if (pLoopUnit->canMoveOrAttackInto(pPlot, bDeclareWar))
			{
				return true;
			}
		}
	}

	return false;
}


bool CvSelectionGroup::canMoveThrough(CvPlot* pPlot)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	if (getNumUnits() > 0)
	{
		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			if (!(pLoopUnit->canMoveThrough(pPlot)))
			{
				return false;
			}
		}

		return true;
	}

	return false;
}


bool CvSelectionGroup::canFight()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if (pLoopUnit->canFight())
		{
			return true;
		}
	}

	return false;
}


bool CvSelectionGroup::canDefend()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if (pLoopUnit->canDefend())
		{
			return true;
		}
	}

	return false;
}


bool CvSelectionGroup::alwaysInvisible()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	if (getNumUnits() > 0)
	{
		pUnitNode = headUnitNode();

		while (pUnitNode != NULL)
		{
			pLoopUnit = ::getUnit(pUnitNode->m_data);
			pUnitNode = nextUnitNode(pUnitNode);

			if (!(pLoopUnit->alwaysInvisible()))
			{
				return false;
			}
		}

		return true;
	}

	return false;
}


int CvSelectionGroup::countNumUnitAIType(UnitAITypes eUnitAI)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	int iCount;

	FAssertMsg(headUnitNode() != NULL, "headUnitNode() is not expected to be equal with NULL");

	iCount = 0;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if (pLoopUnit->AI_getUnitAIType() == eUnitAI)
		{
			iCount++;
		}
	}

	return iCount;
}


bool CvSelectionGroup::hasWorker()
{
	return ((countNumUnitAIType(UNITAI_WORKER) > 0) || (countNumUnitAIType(UNITAI_WORKER_SEA) > 0));
}


bool CvSelectionGroup::IsSelected()
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if (pLoopUnit->IsSelected())
		{
			return true;
		}
	}

	return false;
}


void CvSelectionGroup::NotifyEntity(MissionTypes eMission)
{
	CLLNode<IDInfo>* pUnitNode;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		::getUnit(pUnitNode->m_data)->NotifyEntity(eMission);
		pUnitNode = nextUnitNode(pUnitNode);
	}
}


void CvSelectionGroup::airCircle(bool bStart)
{
	CLLNode<IDInfo>* pUnitNode;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		::getUnit(pUnitNode->m_data)->airCircle(bStart);
		pUnitNode = nextUnitNode(pUnitNode);
	}
}


int CvSelectionGroup::getX() const
{
	CvUnit* pHeadUnit;

	pHeadUnit = getHeadUnit();

	if (pHeadUnit != NULL)
	{
		return getHeadUnit()->getX_INLINE();
	}
	else
	{
		return INVALID_PLOT_COORD;
	}
}


int CvSelectionGroup::getY() const
{
	CvUnit* pHeadUnit;

	pHeadUnit = getHeadUnit();

	if (pHeadUnit != NULL)
	{
		return getHeadUnit()->getY_INLINE();
	}
	else
	{
		return INVALID_PLOT_COORD;
	}
}


bool CvSelectionGroup::at(int iX, int iY) const
{
	return((getX() == iX) && (getY() == iY));
}


bool CvSelectionGroup::atPlot( const CvPlot* pPlot) const
{
	return (plot() == pPlot);
}


CvPlot* CvSelectionGroup::plot() const
{
	CvUnit* pHeadUnit;

	pHeadUnit = getHeadUnit();

	if (pHeadUnit != NULL)
	{
		return getHeadUnit()->plot();
	}
	else
	{
		return NULL;
	}
}


CvArea* CvSelectionGroup::area() const
{
	CvUnit* pHeadUnit;

	pHeadUnit = getHeadUnit();

	if (pHeadUnit != NULL)
	{
		return getHeadUnit()->area();
	}
	else
	{
		return NULL;
	}
}


DomainTypes CvSelectionGroup::getDomainType() const
{
	CvUnit* pHeadUnit;

	pHeadUnit = getHeadUnit();

	if (pHeadUnit != NULL)
	{
		return getHeadUnit()->getDomainType();
	}
	else
	{
		return NO_DOMAIN;
	}
}


RouteTypes CvSelectionGroup::getBestBuildRoute(CvPlot* pPlot, BuildTypes* peBestBuild)
{
	PROFILE_FUNC();

	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	RouteTypes eRoute;
	RouteTypes eBestRoute;
	int iValue;
	int iBestValue;
	int iI;

	if (peBestBuild != NULL)
	{
		*peBestBuild = NO_BUILD;
	}

	iBestValue = 0;
	eBestRoute = NO_ROUTE;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		for (iI = 0; iI < GC.getNumBuildInfos(); iI++)
		{
			eRoute = ((RouteTypes)(GC.getBuildInfo((BuildTypes) iI).getRoute()));

			if (eRoute != NO_ROUTE)
			{
				if (pLoopUnit->canBuild(pPlot, ((BuildTypes)iI)))
				{
					iValue = GC.getRouteInfo(eRoute).getValue();

					if (iValue > iBestValue)
					{
						iBestValue = iValue;
						eBestRoute = eRoute;
						if (peBestBuild != NULL)
						{
							*peBestBuild = ((BuildTypes)iI);
						}
					}
				}
			}
		}
	}

	return eBestRoute;
}


// Returns true if group was bumped...
bool CvSelectionGroup::groupDeclareWar(CvPlot* pPlot)
{
	int iNumUnits;

	if (!AI_isDeclareWar())
	{
		return false;
	}

	iNumUnits = getNumUnits();

	if (!canEnterArea(pPlot->getTeam(), pPlot->area(), (pPlot->getTeam() != NO_TEAM) && GET_TEAM(getTeam()).AI_isSneakAttackReady(pPlot->getTeam())))
	{
		if (GET_TEAM(getTeam()).canDeclareWar(pPlot->getTeam()))
		{
			GET_TEAM(getTeam()).declareWar(pPlot->getTeam(), true);
		}
	}

	return (iNumUnits != getNumUnits());
}


// Returns true if attack was made...
bool CvSelectionGroup::groupAttack(int iX, int iY, int iFlags)
{
	CvUnit* pBestAttackUnit;
	CvPlot* pDestPlot;
	bool bStack;
	bool bAttack;

	pDestPlot = GC.getMapINLINE().plotINLINE(iX, iY);
	FAssertMsg(pDestPlot != NULL, "DestPlot is not assigned a valid value");

	bStack = (isHuman() && ((getDomainType() == DOMAIN_AIR) || GET_PLAYER(getOwnerINLINE()).isOption(PLAYEROPTION_STACK_ATTACK)));

	bAttack = false;

	if (getNumUnits() > 0)
	{
		if ((getDomainType() == DOMAIN_AIR) || (stepDistance(getX(), getY(), pDestPlot->getX_INLINE(), pDestPlot->getY_INLINE()) == 1))
		{
			if ((iFlags & MOVE_DIRECT_ATTACK) || (getDomainType() == DOMAIN_AIR) || (generatePath(plot(), pDestPlot, iFlags) && (getPathFirstPlot() == pDestPlot)))
			{
				pBestAttackUnit = AI_getBestGroupAttacker(pDestPlot, true);

				if (pBestAttackUnit)
				{
					bool bNoBlitz = (!pBestAttackUnit->isBlitz() || !pBestAttackUnit->isMadeAttack());

					if (groupDeclareWar(pDestPlot))
					{
						return true;
					}

					while (true)
					{
						pBestAttackUnit = AI_getBestGroupAttacker(pDestPlot, false, false, bNoBlitz);

						if (pBestAttackUnit == NULL)
						{
							break;
						}

						bAttack = true;

						// Dale - Missiles: missiles code for default attack method START
						if(GC.getUnitInfo(pBestAttackUnit->getUnitType()).getDCMMissile() && DCMMissiles)
						{
							if(GC.getUnitInfo(pBestAttackUnit->getUnitType()).getNukeRange() != -1)
							{
								pBestAttackUnit->nuke(pDestPlot->getX_INLINE(), pDestPlot->getY_INLINE());
								break;
							} else {
								pBestAttackUnit->attack(pDestPlot, bStack);
								pBestAttackUnit->setDamage(pBestAttackUnit->maxHitPoints());
								break;
							}
						}
						// Dale - Missiles: missiles code for default attack method END

						if (getNumUnits() > 1)
						{
							if (pBestAttackUnit->plot()->isFighting() || pDestPlot->isFighting())
							{
								break;
							}
							pBestAttackUnit->attack(pDestPlot, bStack);
						}
						else
						{
							pBestAttackUnit->attack(pDestPlot, false);
							break;
						}

						if (!bStack)
						{
							break;
						}
					}
				}
			}
		}
	}

	return bAttack;
}


void CvSelectionGroup::groupMove(CvPlot* pPlot, bool bCombat, CvUnit* pCombatUnit)
{
	PROFILE_FUNC();

	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if ((pLoopUnit->canMove() && ((bCombat && (!(pLoopUnit->isNoCapture()) || !(pPlot->isEnemyCity(getTeam())))) ? pLoopUnit->canMoveOrAttackInto(pPlot) : pLoopUnit->canMoveInto(pPlot))) || (pLoopUnit == pCombatUnit))
		{
			pLoopUnit->move(pPlot, !bCombat);
		}
		else
		{
			pLoopUnit->joinGroup(NULL, true);
		}
	}
}


// Returns true if move was made...
bool CvSelectionGroup::groupPathTo(int iX, int iY, int iFlags)
{
	CvPlot* pDestPlot;
	CvPlot* pPathPlot;

	if (at(iX, iY))
	{
		return false; // XXX is this necessary?
	}

	FAssert(!isBusy());
	FAssert(getOwnerINLINE() != NO_PLAYER);
	FAssert(headMissionQueueNode() != NULL);

	pDestPlot = GC.getMapINLINE().plotINLINE(iX, iY);
	FAssertMsg(pDestPlot != NULL, "DestPlot is not assigned a valid value");

	FAssertMsg(canAllMove(), "canAllMove is expected to be true");

	if (getDomainType() == DOMAIN_AIR)
	{
		if (!canMoveInto(pDestPlot))
		{
			return false;
		}

		pPathPlot = pDestPlot;
	}
	else
	{
		if (!generatePath(plot(), pDestPlot, iFlags))
		{
			return false;
		}

		pPathPlot = getPathFirstPlot();

		if (groupAmphibMove(pPathPlot, iFlags))
		{
			return false;
		}
	}

	if (groupDeclareWar(pPathPlot))
	{
		return false;
	}

	groupMove(pPathPlot, false);

	return true;
}


// Returns true if move was made...
bool CvSelectionGroup::groupRoadTo(int iX, int iY, int iFlags)
{
	CvPlot* pPlot;
	RouteTypes eBestRoute;
	BuildTypes eBestBuild;

	if (!AI_isControlled() || !at(iX, iY) || (getLengthMissionQueue() == 1))
	{
		pPlot = plot();

		eBestRoute = getBestBuildRoute(pPlot, &eBestBuild);

		if (eBestBuild != NO_BUILD)
		{
			groupBuild(eBestBuild);
			return true;
		}
	}

	return groupPathTo(iX, iY, iFlags);
}


// Returns true if build should continue...
bool CvSelectionGroup::groupBuild(BuildTypes eBuild)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	CvPlot* pPlot;
	bool bContinue;

	FAssert(getOwnerINLINE() != NO_PLAYER);

	bContinue = false;

	pPlot = plot();

	if (GC.getBuildInfo(eBuild).getImprovement() != NO_IMPROVEMENT)
	{
		if (AI_isControlled())
		{
			if (GET_PLAYER(getOwnerINLINE()).isOption(PLAYEROPTION_SAFE_AUTOMATION))
			{
				if (pPlot->getImprovementType() != NO_IMPROVEMENT)
				{
					return false;
				}
			}
		}
	}

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		FAssertMsg(pLoopUnit->atPlot(pPlot), "pLoopUnit is expected to be at pPlot");

		if (pLoopUnit->canBuild(pPlot, eBuild))
		{
			bContinue = true;

			if (pLoopUnit->build(eBuild))
			{
				bContinue = false;
				break;
			}
		}
	}

	return bContinue;
}


bool CvSelectionGroup::isAmphibPlot(CvPlot* pPlot)
{
	return ((getDomainType() == DOMAIN_SEA) && pPlot->isCoastalLand() && !(pPlot->isFriendlyCity(getHeadTeam())));
}


// Returns true if attempted an amphib landing...
bool CvSelectionGroup::groupAmphibMove(CvPlot* pPlot, int iFlags)
{
	CLLNode<IDInfo>* pUnitNode1;
	CLLNode<IDInfo>* pUnitNode2;
	CvUnit* pLoopUnit1;
	CvUnit* pLoopUnit2;

	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (groupDeclareWar(pPlot))
	{
		return true;
	}

	if (isAmphibPlot(pPlot))
	{
		if (stepDistance(getX(), getY(), pPlot->getX_INLINE(), pPlot->getY_INLINE()) == 1)
		{
			pUnitNode1 = headUnitNode();

			while (pUnitNode1 != NULL)
			{
				pLoopUnit1 = ::getUnit(pUnitNode1->m_data);
				pUnitNode1 = nextUnitNode(pUnitNode1);

				if ((pLoopUnit1->getCargo() > 0) && (pLoopUnit1->domainCargo() == DOMAIN_LAND))
				{
					pUnitNode2 = plot()->headUnitNode();

					while (pUnitNode2 != NULL)
					{
						pLoopUnit2 = ::getUnit(pUnitNode2->m_data);
						pUnitNode2 = plot()->nextUnitNode(pUnitNode2);

						if (pLoopUnit2->getTransportUnit() == pLoopUnit1)
						{
							if (pLoopUnit2->isGroupHead())
							{
								pLoopUnit2->getGroup()->pushMission(MISSION_MOVE_TO, pPlot->getX_INLINE(), pPlot->getY_INLINE(), (MOVE_IGNORE_DANGER | iFlags));
							}
						}
					}
				}
			}

			return true;
		}
	}

	return false;
}


bool CvSelectionGroup::readyToSelect(bool bAny)
{
	return (readyToMove(bAny) && !isAutomated());
}


bool CvSelectionGroup::readyToMove(bool bAny)
{
	return (((bAny) ? canAnyMove() : canAllMove()) && (headMissionQueueNode() == NULL) && (getActivityType() == ACTIVITY_AWAKE) && !isBusy() && !isCargoBusy());
}


bool CvSelectionGroup::readyToAuto()
{
	return (canAllMove() && (headMissionQueueNode() != NULL));
}


int CvSelectionGroup::getID()
{
	return m_iID;
}


void CvSelectionGroup::setID(int iID)																			
{
	m_iID = iID;
}


PlayerTypes CvSelectionGroup::getOwner() const
{
	return getOwnerINLINE();
}


TeamTypes CvSelectionGroup::getTeam() const
{
	if (getOwnerINLINE() != NO_PLAYER)
	{
		return GET_PLAYER(getOwnerINLINE()).getTeam();
	}

	return NO_TEAM;
}


int CvSelectionGroup::getMissionTimer()
{
	return m_iMissionTimer;
}


void CvSelectionGroup::setMissionTimer(int iNewValue)
{
	FAssert(getOwnerINLINE() != NO_PLAYER);

	m_iMissionTimer = iNewValue;
	FAssert(getMissionTimer() >= 0);
}


void CvSelectionGroup::changeMissionTimer(int iChange)
{
	setMissionTimer(getMissionTimer() + iChange);
}


void CvSelectionGroup::updateMissionTimer(int iSteps)
{
	CvUnit* pTargetUnit;
	CvPlot* pTargetPlot;
	int iTime;

	if (!isHuman() && !(GC.getGameINLINE().showAllMoves(getTeam())))
	{
		iTime = 0;
	}
	else if (headMissionQueueNode() != NULL)
	{
		iTime = GC.getMissionInfo((MissionTypes)(headMissionQueueNode()->m_data.eMissionType)).getTime();

		if ((headMissionQueueNode()->m_data.eMissionType == MISSION_MOVE_TO) ||
				(headMissionQueueNode()->m_data.eMissionType == MISSION_ROUTE_TO) ||
				(headMissionQueueNode()->m_data.eMissionType == MISSION_MOVE_TO_UNIT))
		{
			if (headMissionQueueNode()->m_data.eMissionType == MISSION_MOVE_TO_UNIT)
			{
				pTargetUnit = GET_PLAYER((PlayerTypes)headMissionQueueNode()->m_data.iData1).getUnit(headMissionQueueNode()->m_data.iData2);
				if (pTargetUnit != NULL)
				{
					pTargetPlot = pTargetUnit->plot();
				}
				else
				{
					pTargetPlot = NULL;
				}
			}
			else
			{
				pTargetPlot = GC.getMapINLINE().plotINLINE(headMissionQueueNode()->m_data.iData1, headMissionQueueNode()->m_data.iData2);
			}

			if (atPlot(pTargetPlot))
			{
				iTime += iSteps;
			}
			else
			{
				iTime = min(iTime, 2);
			}
		}

		if (isHuman() && (isAutomated() || (GET_PLAYER((GC.getGameINLINE().isNetworkMultiPlayer()) ? getOwnerINLINE() : GC.getGameINLINE().getActivePlayer()).isOption(PLAYEROPTION_QUICK_MOVES))))
		{
			iTime = min(iTime, 1);
		}
	}
	else
	{
		iTime = 0;
	}

	setMissionTimer(iTime);
}


bool CvSelectionGroup::isForceUpdate()
{
	return m_bForceUpdate;
}


void CvSelectionGroup::setForceUpdate(bool bNewValue)
{
	m_bForceUpdate = bNewValue;
}


ActivityTypes CvSelectionGroup::getActivityType() const
{
	return m_eActivityType;
}


void CvSelectionGroup::setActivityType(ActivityTypes eNewValue)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	CvPlot* pPlot;

	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (getActivityType() != eNewValue)
	{
		pPlot = plot();

		if (getActivityType() == ACTIVITY_INTERCEPT)
		{
			airCircle(false);
		}

		m_eActivityType = eNewValue;

		if (getActivityType() == ACTIVITY_INTERCEPT)
		{
			airCircle(true);
		}

		if (getActivityType() != ACTIVITY_MISSION)
		{
			pUnitNode = headUnitNode();

			while (pUnitNode != NULL)
			{
				pLoopUnit = ::getUnit(pUnitNode->m_data);
				pUnitNode = nextUnitNode(pUnitNode);

				pLoopUnit->NotifyEntity(MISSION_IDLE);
			}

			if (getTeam() == GC.getGameINLINE().getActiveTeam())
			{
				if (pPlot != NULL)
				{
					pPlot->setFlagDirty(true);
				}
			}
		}

		if (pPlot == gDLL->getInterfaceIFace()->getSelectionPlot())
		{
			gDLL->getInterfaceIFace()->setDirty(PlotListButtons_DIRTY_BIT, true);
			gDLL->getInterfaceIFace()->setDirty(SelectionButtons_DIRTY_BIT, true);
		}
	}
}


AutomateTypes CvSelectionGroup::getAutomateType()															
{
	return m_eAutomateType;
}


bool CvSelectionGroup::isAutomated()
{
	return (getAutomateType() != NO_AUTOMATE);
}


void CvSelectionGroup::setAutomateType(AutomateTypes eNewValue)
{
	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (getAutomateType() != eNewValue)
	{
		m_eAutomateType = eNewValue;

		clearMissionQueue();
		setActivityType(ACTIVITY_AWAKE);
	}
}


FAStarNode* CvSelectionGroup::getPathLastNode() const
{
	return gDLL->getFAStarIFace()->GetLastNode(&GC.getPathFinder());
}


CvPlot* CvSelectionGroup::getPathFirstPlot() const
{
	FAStarNode* pNode;

	pNode = getPathLastNode();

	if (pNode->m_pParent == NULL)
	{
		return GC.getMapINLINE().plotSorenINLINE(pNode->m_iX, pNode->m_iY);
	}

	while (pNode != NULL)
	{
		if (pNode->m_pParent->m_pParent == NULL)
		{
			return GC.getMapINLINE().plotSorenINLINE(pNode->m_iX, pNode->m_iY);
		}

		pNode = pNode->m_pParent;
	}

	FAssert(false);

	return NULL;
}


CvPlot* CvSelectionGroup::getPathEndTurnPlot() const
{
	FAStarNode* pNode;

	pNode = getPathLastNode();

	if ((pNode->m_pParent == NULL) || (pNode->m_iData2 == 1))
	{
		return GC.getMapINLINE().plotSorenINLINE(pNode->m_iX, pNode->m_iY);
	}

	while (pNode->m_pParent != NULL)
	{
		if (pNode->m_pParent->m_iData2 == 1)
		{
			return GC.getMapINLINE().plotSorenINLINE(pNode->m_pParent->m_iX, pNode->m_pParent->m_iY);
		}

		pNode = pNode->m_pParent;
	}

	FAssert(false);

	return NULL;
}


bool CvSelectionGroup::generatePath( const CvPlot* pFromPlot, const CvPlot* pToPlot, int iFlags, bool bReuse, int* piPathTurns) const
{
	PROFILE("CvSelectionGroup::generatePath()")

	FAStarNode* pNode;
	bool bSuccess;

	gDLL->getFAStarIFace()->SetData(&GC.getPathFinder(), this);

	bSuccess = gDLL->getFAStarIFace()->GeneratePath(&GC.getPathFinder(), pFromPlot->getX_INLINE(), pFromPlot->getY_INLINE(), pToPlot->getX_INLINE(), pToPlot->getY_INLINE(), false, iFlags, bReuse);

	if (piPathTurns != NULL)
	{
		*piPathTurns = MAX_INT;

		if (bSuccess)
		{
			pNode = getPathLastNode();

			if (pNode != NULL)
			{
				*piPathTurns = pNode->m_iData2;
			}
		}
	}

	return bSuccess;
}


void CvSelectionGroup::resetPath()
{
	gDLL->getFAStarIFace()->ForceReset(&GC.getPathFinder());
}


void CvSelectionGroup::clearUnits()
{
	CLLNode<IDInfo>* pUnitNode;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pUnitNode = deleteUnitNode(pUnitNode);
	}
}


// Returns true if the unit is added...
bool CvSelectionGroup::addUnit(CvUnit* pUnit)
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	bool bAdded;
 
	if (!(pUnit->canJoinGroup(pUnit->plot(), this)))
	{
		return false;
	}

	bAdded = false;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		if ((pUnit->AI_groupFirstVal() > pLoopUnit->AI_groupFirstVal()) ||
			  ((pUnit->AI_groupFirstVal() == pLoopUnit->AI_groupFirstVal()) &&
				 (pUnit->AI_groupSecondVal() > pLoopUnit->AI_groupSecondVal())))
		{
			m_units.insertBefore(pUnit->getIDInfo(), pUnitNode);
			bAdded = true;
			break;
		}
		pUnitNode = nextUnitNode(pUnitNode);
	}

	if (!bAdded)
	{
		m_units.insertAtEnd(pUnit->getIDInfo());
	}

	if (getOwnerINLINE() == NO_PLAYER)
	{
		if (getNumUnits() > 0)
		{
			pUnitNode = headUnitNode();
			while (pUnitNode != NULL)
			{
				if (pUnitNode != headUnitNode())
				{
					::getUnit(pUnitNode->m_data)->NotifyEntity(MISSION_MULTI_SELECT);
				}
				pUnitNode = nextUnitNode(pUnitNode);
			}
		}
	}

	return true;
}


void CvSelectionGroup::removeUnit(CvUnit* pUnit)
{
	CLLNode<IDInfo>* pUnitNode;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		if (::getUnit(pUnitNode->m_data) == pUnit)
		{
			deleteUnitNode(pUnitNode);
			break;
		}
		else
		{
			pUnitNode = nextUnitNode(pUnitNode);
		}
	}
}


CLLNode<IDInfo>* CvSelectionGroup::deleteUnitNode(CLLNode<IDInfo>* pNode)
{
	CLLNode<IDInfo>* pNextUnitNode;

	if (getOwnerINLINE() != NO_PLAYER)
	{
		setAutomateType(NO_AUTOMATE);
		clearMissionQueue();
		if (getActivityType() != ACTIVITY_SLEEP)
		{
			setActivityType(ACTIVITY_AWAKE);
		}
	}

	pNextUnitNode = m_units.deleteNode(pNode);

	return pNextUnitNode;
}


CLLNode<IDInfo>* CvSelectionGroup::nextUnitNode(CLLNode<IDInfo>* pNode) const
{
	return m_units.next(pNode);
}


int CvSelectionGroup::getNumUnits() const
{
	return m_units.getLength();
}

//------------------------------------------------------------------------------------------------
// FUNCTION:    CvSelectionGroup::getUnitIndex
//! \brief      Returns the index of the given unit in the selection group
//! \param      pUnit The unit to find the index of within the group
//! \retval     The zero-based index of the unit within the group, or -1 if it is not in the group.
//------------------------------------------------------------------------------------------------
int CvSelectionGroup::getUnitIndex(CvUnit* pUnit) const
{
	CLLNode<IDInfo>* pUnitNode;
	CvUnit* pLoopUnit;
	int iIndex;

	iIndex = 0;

	pUnitNode = headUnitNode();

	while (pUnitNode != NULL)
	{
		pLoopUnit = ::getUnit(pUnitNode->m_data);
		pUnitNode = nextUnitNode(pUnitNode);

		if (pLoopUnit == pUnit)
		{
			return iIndex;
		}

		iIndex++;
	}

	return -1;
}

CLLNode<IDInfo>* CvSelectionGroup::headUnitNode() const
{
	return m_units.head();
}


CvUnit* CvSelectionGroup::getHeadUnit() const
{
	CLLNode<IDInfo>* pUnitNode;

	pUnitNode = headUnitNode();

	if (pUnitNode != NULL)
	{
		return ::getUnit(pUnitNode->m_data);
	}
	else
	{
		return NULL;
	}
}


UnitAITypes CvSelectionGroup::getHeadUnitAI() const
{
	CvUnit* pHeadUnit;

	pHeadUnit = getHeadUnit();

	if (pHeadUnit != NULL)
	{
		return pHeadUnit->AI_getUnitAIType();
	}

	return NO_UNITAI;
}


PlayerTypes CvSelectionGroup::getHeadOwner() const
{
	CvUnit* pHeadUnit;

	pHeadUnit = getHeadUnit();

	if (pHeadUnit != NULL)
	{
		return pHeadUnit->getOwnerINLINE();
	}

	return NO_PLAYER;
}


TeamTypes CvSelectionGroup::getHeadTeam() const
{
	CvUnit* pHeadUnit;

	pHeadUnit = getHeadUnit();

	if (pHeadUnit != NULL)
	{
		return pHeadUnit->getTeam();
	}

	return NO_TEAM;
}


void CvSelectionGroup::clearMissionQueue()
{
	FAssert(getOwnerINLINE() != NO_PLAYER);

	deactivateHeadMission();

	m_missionQueue.clear();

	if ((getOwnerINLINE() == GC.getGameINLINE().getActivePlayer()) && IsSelected())
	{
		gDLL->getInterfaceIFace()->setDirty(Waypoints_DIRTY_BIT, true);
		gDLL->getInterfaceIFace()->setDirty(SelectionButtons_DIRTY_BIT, true);
		gDLL->getInterfaceIFace()->setDirty(InfoPane_DIRTY_BIT, true);
	}
}


int CvSelectionGroup::getLengthMissionQueue()																	
{
	return m_missionQueue.getLength();
}


MissionData* CvSelectionGroup::getMissionFromQueue(int iIndex)
{
	CLLNode<MissionData>* pMissionNode;

	pMissionNode = m_missionQueue.nodeNum(iIndex);

	if (pMissionNode != NULL)
	{
		return &(pMissionNode->m_data);
	}
	else
	{
		return NULL;
	}
}


void CvSelectionGroup::insertAtEndMissionQueue(MissionData mission, bool bStart)
{
	FAssert(getOwnerINLINE() != NO_PLAYER);

	m_missionQueue.insertAtEnd(mission);

	if ((getLengthMissionQueue() == 1) && bStart)
	{
		activateHeadMission();
	}

	if ((getOwnerINLINE() == GC.getGameINLINE().getActivePlayer()) && IsSelected())
	{
		gDLL->getInterfaceIFace()->setDirty(Waypoints_DIRTY_BIT, true);
		gDLL->getInterfaceIFace()->setDirty(SelectionButtons_DIRTY_BIT, true);
		gDLL->getInterfaceIFace()->setDirty(InfoPane_DIRTY_BIT, true);
	}
}


CLLNode<MissionData>* CvSelectionGroup::deleteMissionQueueNode(CLLNode<MissionData>* pNode)
{
	CLLNode<MissionData>* pNextMissionNode;

	FAssertMsg(pNode != NULL, "Node is not assigned a valid value");
	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (pNode == headMissionQueueNode())
	{
		deactivateHeadMission();
	}

	pNextMissionNode = m_missionQueue.deleteNode(pNode);

	if (pNextMissionNode == headMissionQueueNode())
	{
		activateHeadMission();
	}

	if ((getOwnerINLINE() == GC.getGameINLINE().getActivePlayer()) && IsSelected())
	{
		gDLL->getInterfaceIFace()->setDirty(Waypoints_DIRTY_BIT, true);
		gDLL->getInterfaceIFace()->setDirty(SelectionButtons_DIRTY_BIT, true);
		gDLL->getInterfaceIFace()->setDirty(InfoPane_DIRTY_BIT, true);
	}

	return pNextMissionNode;
}


CLLNode<MissionData>* CvSelectionGroup::nextMissionQueueNode(CLLNode<MissionData>* pNode) const
{
	return m_missionQueue.next(pNode);
}


CLLNode<MissionData>* CvSelectionGroup::prevMissionQueueNode(CLLNode<MissionData>* pNode) const
{
	return m_missionQueue.prev(pNode);
}


CLLNode<MissionData>* CvSelectionGroup::headMissionQueueNode() const
{
	return m_missionQueue.head();
}


CLLNode<MissionData>* CvSelectionGroup::tailMissionQueueNode() const
{
	return m_missionQueue.tail();
}


int CvSelectionGroup::getMissionType(int iNode)
{
	int iCount = 0;
	CLLNode<MissionData>* pMissionNode;

	pMissionNode = headMissionQueueNode();

	while (pMissionNode != NULL)
	{
		if ( iNode == iCount )
		{
			return pMissionNode->m_data.eMissionType;
		}

		iCount++;

		pMissionNode = nextMissionQueueNode(pMissionNode);
	}

	return -1;
}


int CvSelectionGroup::getMissionData1(int iNode)
{
	int iCount = 0;
	CLLNode<MissionData>* pMissionNode;

	pMissionNode = headMissionQueueNode();

	while (pMissionNode != NULL)
	{
		if ( iNode == iCount )
		{
			return pMissionNode->m_data.iData1;
		}

		iCount++;

		pMissionNode = nextMissionQueueNode(pMissionNode);
	}

	return -1;
}


int CvSelectionGroup::getMissionData2(int iNode)
{
	int iCount = 0;
	CLLNode<MissionData>* pMissionNode;

	pMissionNode = headMissionQueueNode();

	while (pMissionNode != NULL)
	{
		if ( iNode == iCount )
		{
			return pMissionNode->m_data.iData2;
		}

		iCount++;

		pMissionNode = nextMissionQueueNode(pMissionNode);
	}

	return -1;
}


void CvSelectionGroup::read(FDataStreamBase* pStream)
{
	// Init saved data
	reset();

	uint uiFlag=0;
	pStream->Read(&uiFlag);	// flags for expansion

	pStream->Read(&m_iID);
	pStream->Read(&m_iMissionTimer);

	pStream->Read(&m_bForceUpdate);

	pStream->Read((int*)&m_eOwner);
	pStream->Read((int*)&m_eActivityType);
	pStream->Read((int*)&m_eAutomateType);

	m_units.Read(pStream);
	m_missionQueue.Read(pStream);
}


void CvSelectionGroup::write(FDataStreamBase* pStream)
{
	uint uiFlag=0;
	pStream->Write(uiFlag);		// flag for expansion

	pStream->Write(m_iID);
	pStream->Write(m_iMissionTimer);

	pStream->Write(m_bForceUpdate);

	pStream->Write(m_eOwner);
	pStream->Write(m_eActivityType);
	pStream->Write(m_eAutomateType);

	m_units.Write(pStream);
	m_missionQueue.Write(pStream);
}

// Protected Functions...

void CvSelectionGroup::activateHeadMission()
{
	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (headMissionQueueNode() != NULL)
	{
		if (!isBusy())
		{
			startMission();
		}
	}
}


void CvSelectionGroup::deactivateHeadMission()
{
	FAssert(getOwnerINLINE() != NO_PLAYER);

	if (headMissionQueueNode() != NULL)
	{
		if (getActivityType() == ACTIVITY_MISSION)
		{
			setActivityType(ACTIVITY_AWAKE);
		}

		setMissionTimer(0);

		if (getOwnerINLINE() == GC.getGameINLINE().getActivePlayer())
		{
			if (IsSelected())
			{
				gDLL->getInterfaceIFace()->changeCycleSelectionCounter(1);
			}
		}
	}
}

// Dale - Stack Attack: new groupStackAttack() function
bool CvSelectionGroup::groupStackAttack(int iX, int iY, int iFlags)
{
//	CvUnit* pBestAttackUnit;
	CvPlot* pDestPlot;
	bool bStack;
	bool bAttack;

	CvPlot* pOrigPlot;				// Dale - Stack Attack: origin plot of attack
	CvUnit* uAttLine[4][36];		// Dale - Stack Attack: attacker lines
	CvUnit* uDefLine[4][36];		// Dale - Stack Attack: defender lines
	int iNumRows = 4;				// Dale - Stack Attack: number of rows per side
	int iNumCols = 36;				// Dale - Stack Attack: number of columns per row
	int iCount, iCount2, iCount3, iIndex, iIndex2;	// Dale - Stack Attack: required ints
	int iNumAtts, iNumDefs;			// Dale - Stack Attack: number of attackers & defenders remaining
	int iAttsRow[4], iDefsRow[4];	// Dale - Stack Attack: number of units in each row
	CLLNode<IDInfo>* pUnitNode;		// Dale - Stack Attack: list
	CLLNode<IDInfo>* pUnitNode2;	// Dale - Stack Attack: list
	CvUnit* pLoopUnit;				// Dale - Stack Attack: list
	CvUnit* pLoopUnit2;				// Dale - Stack Attack: list
	int iAttForm = 1;				// Dale - Stack Attack: attacker formation
	int iDefForm = 1;				// Dale - Stack Attack: defender formation
	CvSelectionGroup* gAttack;		// Dale - Stack Attack: required for AI attack

	pOrigPlot = plot();	// Dale - Stack Attack: set origin plot
	FAssertMsg(pOrigPlot != NULL, "OrigPlot is not assigned a valid value");
	pDestPlot = GC.getMapINLINE().plotINLINE(iX, iY);
	FAssertMsg(pDestPlot != NULL, "DestPlot is not assigned a valid value");

	bStack = true;//(isHuman() && ((getDomainType() == DOMAIN_AIR) || GET_PLAYER(getOwnerINLINE()).isOption(PLAYEROPTION_STACK_ATTACK)));

	bAttack = false;

	if (getNumUnits() > 0)
	{
		if ((getDomainType() == DOMAIN_AIR) || (stepDistance(getX(), getY(), pDestPlot->getX_INLINE(), pDestPlot->getY_INLINE()) == 1))
		{
			if ((iFlags & MOVE_DIRECT_ATTACK) || (getDomainType() == DOMAIN_AIR) || (generatePath(plot(), pDestPlot, iFlags) && (getPathFirstPlot() == pDestPlot)))
			{
				if (AI_getBestGroupAttacker(pDestPlot, true) != NULL)
				{
					if (groupDeclareWar(pDestPlot))
					{
						return true;
					}

					// Dale - Stack Attack: reset attacker and defender lines
					for(iCount = 0; iCount < iNumRows; iCount++)
					{
						iAttsRow[iCount] = 0;
						iDefsRow[iCount] = 0;
						for(iCount2 = 0; iCount2 < iNumCols; iCount2++)
						{
							uAttLine[iCount][iCount2] = NULL;
							uDefLine[iCount][iCount2] = NULL;
						}
					}

					// Dale - Stack Attack: setup attackers
					if(isHuman())
					{
                        pUnitNode = headUnitNode();
					} else {
						// Dale - Stack Attack: group AI units and weed out non-combat units
						gAttack = NULL;
						pUnitNode = pOrigPlot->headUnitNode();
						while(pUnitNode != NULL)
						{
                            pLoopUnit2 = ::getUnit(pUnitNode->m_data);
                            pUnitNode = nextUnitNode(pUnitNode);
							if(pLoopUnit2->getDomainType() == DOMAIN_AIR || (pDestPlot->isWater() && pLoopUnit2->getDomainType() == DOMAIN_SEA) || (!pDestPlot->isWater() && !pLoopUnit2->getDomainType() == DOMAIN_SEA))
							{
								if(pLoopUnit2->canAttack() && pLoopUnit2->canMove())
								{
									gAttack = pLoopUnit2->getGroup();
									pUnitNode = NULL;
								}
							}
						}
						if(gAttack == NULL)
						{
							return bAttack;
						}
						pUnitNode = pOrigPlot->headUnitNode();
						while(pUnitNode != NULL)
						{
							pLoopUnit = ::getUnit(pUnitNode->m_data);
							pUnitNode = nextUnitNode(pUnitNode);
							if(pLoopUnit->canAttack() && pLoopUnit != pLoopUnit2 && pLoopUnit->canMove())
							{
								pLoopUnit->joinGroup(gAttack);
							}
						}
						if(gAttack != NULL)
						{
							pUnitNode = gAttack->headUnitNode();
						} else {
							return bAttack;
						}
					}
					// Dale - Stack Attack: form up attacker
					while (pUnitNode != NULL)
					{
						iNumAtts = 0;
						pLoopUnit = ::getUnit(pUnitNode->m_data);
						pUnitNode = nextUnitNode(pUnitNode);
						// Dale - Missiles: section that performs the missile nuke stuff
						if((GC.getUnitInfo(pLoopUnit->getUnitType()).getDCMMissile() && DCMMissiles) || (getNumUnits() == 1 && getDomainType() == DOMAIN_AIR))
						{
							if(GC.getUnitInfo(pLoopUnit->getUnitType()).getDCMMissile())
							{
								if(GC.getUnitInfo(pLoopUnit->getUnitType()).getNukeRange() != -1)
								{
									pLoopUnit->nuke(pDestPlot->getX_INLINE(), pDestPlot->getY_INLINE());
								} else {
									pLoopUnit->updateStackCombat(pDestPlot->getBestDefender(NO_PLAYER, pLoopUnit->getOwnerINLINE(), pLoopUnit, true), 2, false, bStack);
                                    pLoopUnit->setDamage(pLoopUnit->maxHitPoints());
								}
							} else {
								pLoopUnit->updateStackCombat(pDestPlot->getBestDefender(NO_PLAYER, pLoopUnit->getOwnerINLINE(), pLoopUnit, true), 2, false, bStack);
							}
						} else if(!(pLoopUnit->isDead()) && pLoopUnit->canAttack()) {
							pLoopUnit->setDCMAttDef(2);
							switch(iAttForm)
							{
							case 1:
								if(pLoopUnit->getDomainType() == DOMAIN_AIR)
								{
									iIndex = 0;
									iCount = 0;
								} else if(pLoopUnit->AI_getUnitAIType() == UNITAI_WORKER || pLoopUnit->AI_getUnitAIType() == UNITAI_WORKER_SEA) {
									iIndex = 1;
									iCount = 1;
								} else if(pLoopUnit->AI_getUnitAIType() == UNITAI_SETTLE || pLoopUnit->AI_getUnitAIType() == UNITAI_SETTLER_SEA) {
									iIndex = 1;
									iCount = 1;
								} else if(pLoopUnit->AI_getUnitAIType() == UNITAI_MISSIONARY || pLoopUnit->AI_getUnitAIType() == UNITAI_PROPHET || pLoopUnit->AI_getUnitAIType() == UNITAI_ARTIST || pLoopUnit->AI_getUnitAIType() == UNITAI_SCIENTIST || pLoopUnit->AI_getUnitAIType() == UNITAI_MERCHANT || pLoopUnit->AI_getUnitAIType() == UNITAI_ENGINEER || pLoopUnit->AI_getUnitAIType() == UNITAI_SPY || pLoopUnit->AI_getUnitAIType() == UNITAI_MISSIONARY_SEA || pLoopUnit->AI_getUnitAIType() == UNITAI_SPY_SEA) {
									iIndex = 1;
									iCount = 1;
								} else if(pLoopUnit->getDCMBombard()) {
									iIndex = 1;
									iCount = 1;
								} else if(pLoopUnit->getDCMRanged()) {
									iIndex = 2;
									iCount = 2;
								} else if(pLoopUnit->getDCMFlanker()) {
									iIndex = 3;
									iCount = 3;
								} else {
									iIndex = 4;
									iCount = 3;
								}
								iIndex2 = (iNumCols / 2);
								if(iIndex == 3)
								{
									iCount2 = (iIndex2 - 2);
								} else {
									iCount2 = 0;
								}
								if(iAttsRow[iCount] <= 20 && iNumAtts <= 30)
								{
									while(true)
									{
										if(uAttLine[iCount][(iIndex2 - iCount2)] == NULL)
										{
											uAttLine[iCount][(iIndex2 - iCount2)] = pLoopUnit;
											iAttsRow[iCount]++;
											iNumAtts++;
											break;
										} else if(uAttLine[iCount][(iIndex2 + iCount2)] == NULL) {
											uAttLine[iCount][(iIndex2 + iCount2)] = pLoopUnit;
											iAttsRow[iCount]++;
											iNumAtts++;
											break;
										}
										if(iIndex == 3)
										{
											iCount2--;
											if(iCount2 == 2)
											{
												break;
											}
										} else {
											iCount2++;
											if(iCount2 == (iIndex2 - 2))
											{
												break;
											}
										}
									}
									if(pLoopUnit->getCargo() > 0)
									{
										pUnitNode2 = pOrigPlot->headUnitNode();
										while(pUnitNode2 != NULL)
										{
											pLoopUnit2 = ::getUnit(pUnitNode2->m_data);
											pUnitNode2 = nextUnitNode(pUnitNode2);
											if(pLoopUnit2->getTransportUnit() == pLoopUnit)
											{
												if(pLoopUnit2->getDomainType() == DOMAIN_AIR)
												{
													iCount3 = 0;
													while(true)
													{
														if(uAttLine[0][(iIndex2 - iCount3)] == NULL)
														{
															uAttLine[0][(iIndex2 - iCount3)] = pLoopUnit2;
															iAttsRow[0]++;
															iNumAtts++;
															break;
														} else if(uAttLine[0][(iIndex2 + iCount3)] == NULL) {
															uAttLine[0][(iIndex2 + iCount3)] = pLoopUnit2;
															iAttsRow[0]++;
															iNumAtts++;
															break;
														}
														iCount3++;
														if(iCount3 == (iIndex2 - 2))
														{
															break;
														}
													}
												}
											}
										}
									}
								}
								break;
							}
						}
					}

					// Dale - Stack Attack: setup defenders
					pUnitNode = pDestPlot->headUnitNode();
					while (pUnitNode != NULL)
					{
						iNumDefs = 0;
						pLoopUnit = ::getUnit(pUnitNode->m_data);
						pUnitNode = nextUnitNode(pUnitNode);
						if(atWar(getTeam(), pLoopUnit->getTeam()))
						{
							if(!(pLoopUnit->isDead()))
							{
								pLoopUnit->setDCMAttDef(1);
								if(!pLoopUnit->isHuman() && !isHuman())
								{
									bStack = false;
								}
								switch(iDefForm)
								{
								case 1:
									if(pLoopUnit->getDomainType() == DOMAIN_AIR)
									{
										iIndex = 0;
										iCount = 0;
									} else if(pLoopUnit->AI_getUnitAIType() == UNITAI_WORKER || pLoopUnit->AI_getUnitAIType() == UNITAI_WORKER_SEA) {
										iIndex = 1;
										iCount = 1;
									} else if(pLoopUnit->AI_getUnitAIType() == UNITAI_SETTLE || pLoopUnit->AI_getUnitAIType() == UNITAI_SETTLER_SEA) {
										iIndex = 1;
										iCount = 1;
									} else if(pLoopUnit->AI_getUnitAIType() == UNITAI_MISSIONARY || pLoopUnit->AI_getUnitAIType() == UNITAI_PROPHET || pLoopUnit->AI_getUnitAIType() == UNITAI_ARTIST || pLoopUnit->AI_getUnitAIType() == UNITAI_SCIENTIST || pLoopUnit->AI_getUnitAIType() == UNITAI_MERCHANT || pLoopUnit->AI_getUnitAIType() == UNITAI_ENGINEER || pLoopUnit->AI_getUnitAIType() == UNITAI_SPY || pLoopUnit->AI_getUnitAIType() == UNITAI_MISSIONARY_SEA || pLoopUnit->AI_getUnitAIType() == UNITAI_SPY_SEA) {
										iIndex = 1;
										iCount = 1;
									} else if(pLoopUnit->getDCMBombard()) {
										iIndex = 1;
										iCount = 1;
									} else if(pLoopUnit->getDCMRanged()) {
										iIndex = 2;
										iCount = 2;
									} else if(pLoopUnit->getDCMFlanker()) {
										iIndex = 3;
										iCount = 3;
									} else {
										iIndex = 4;
										iCount = 3;
									}
									iIndex2 = (iNumCols / 2);
									if(iIndex == 3)
									{
										iCount2 = (iIndex2 - 2);
									} else {
										iCount2 = 0;
									}
									if(iDefsRow[iCount] <= 20 && iNumDefs <= 30)
									{
										while(true)
										{
											if(uDefLine[iCount][(iIndex2 - iCount2)] == NULL)
											{
												uDefLine[iCount][(iIndex2 - iCount2)] = pLoopUnit;
												iDefsRow[iCount]++;
												iNumDefs++;
												break;
											} else if(uDefLine[iCount][(iIndex2 + iCount2)] == NULL) {
												uDefLine[iCount][(iIndex2 + iCount2)] = pLoopUnit;
												iDefsRow[iCount]++;
												iNumDefs++;
												break;
											}
											if(iIndex == 3)
											{
												iCount2--;
												if(iCount2 == 2)
												{
													break;
												}
											} else {
												iCount2++;
												if(iCount2 == (iIndex2 - 2))
												{
													break;
												}
											}
										}
									}
									break;
								}
							}
						}
					}

					// Dale - Stack Attack: unit shuffler
					iIndex = (iNumCols / 2);
					iIndex2 = 1;
					while(iIndex2 > 0)
					{
						iIndex2 = 0;
						for(iCount = 0; iCount < iNumRows; iCount++)
						{
							for(iCount2 = 0; iCount2 < (iIndex - 2); iCount2++)
							{
								if(uAttLine[iCount][(iIndex - iCount2)] == NULL && uAttLine[iCount][(iIndex - (iCount2 + 1))] != NULL)
								{
									uAttLine[iCount][(iIndex - iCount2)] = uAttLine[iCount][(iIndex - (iCount2 + 1))];
									uAttLine[iCount][(iIndex - (iCount2 + 1))] = NULL;
									iIndex2++;
								}
								if(uAttLine[iCount][(iIndex + iCount2)] == NULL && uAttLine[iCount][(iIndex + (iCount2 + 1))] != NULL)
								{
									uAttLine[iCount][(iIndex + iCount2)] = uAttLine[iCount][(iIndex + (iCount2 + 1))];
									uAttLine[iCount][(iIndex + (iCount2 + 1))] = NULL;
									iIndex2++;
								}
								if(uDefLine[iCount][(iIndex - iCount2)] == NULL && uDefLine[iCount][(iIndex - (iCount2 + 1))] != NULL)
								{
									uDefLine[iCount][(iIndex - iCount2)] = uDefLine[iCount][(iIndex - (iCount2 + 1))];
									uDefLine[iCount][(iIndex - (iCount2 + 1))] = NULL;
									iIndex2++;
								}
								if(uDefLine[iCount][(iIndex + iCount2)] == NULL && uDefLine[iCount][(iIndex + (iCount2 + 1))] != NULL)
								{
									uDefLine[iCount][(iIndex + iCount2)] = uDefLine[iCount][(iIndex + (iCount2 + 1))];
									uDefLine[iCount][(iIndex + (iCount2 + 1))] = NULL;
									iIndex2++;
								}
							}
						}
					}

					// Dale - Stack Attack: Combat rounds

					// Dale - Stack Attack: defender airstrike
					for(iCount = 2; iCount < (iNumCols - 3); iCount++)
					{
						if(uDefLine[0][iCount] != NULL)
						{
							if(!uDefLine[0][iCount]->isDead())
							{
								if(uAttLine[0][iCount] != NULL)
								{
									if(!uAttLine[0][iCount]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[0][(iCount - 1)] != NULL) {
									if(!uAttLine[0][(iCount - 1)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][(iCount - 1)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[0][(iCount + 1)] != NULL) {
									if(!uAttLine[0][(iCount + 1)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][(iCount + 1)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[0][(iCount - 2)] != NULL) {
									if(!uAttLine[0][(iCount - 2)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][(iCount - 2)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[0][(iCount + 2)] != NULL) {
									if(!uAttLine[0][(iCount + 2)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][(iCount + 2)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[1][iCount] != NULL) {
									if(!uAttLine[1][iCount]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[1][iCount]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[1][(iCount - 1)] != NULL) {
									if(!uAttLine[1][(iCount - 1)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[1][(iCount - 1)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[1][(iCount + 1)] != NULL) {
									if(!uAttLine[1][(iCount + 1)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[1][(iCount + 1)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[1][(iCount - 2)] != NULL) {
									if(!uAttLine[1][(iCount - 2)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[1][(iCount - 2)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[1][(iCount + 2)] != NULL) {
									if(!uAttLine[1][(iCount + 2)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[1][(iCount + 2)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[2][iCount] != NULL) {
									if(!uAttLine[2][iCount]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[2][iCount]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[2][(iCount - 1)] != NULL) {
									if(!uAttLine[2][(iCount - 1)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[2][(iCount - 1)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[2][(iCount + 1)] != NULL) {
									if(!uAttLine[2][(iCount + 1)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[2][(iCount + 1)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[2][(iCount - 2)] != NULL) {
									if(!uAttLine[2][(iCount - 2)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[2][(iCount - 2)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[2][(iCount + 2)] != NULL) {
									if(!uAttLine[2][(iCount + 2)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[2][(iCount + 2)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[3][iCount] != NULL) {
									if(!uAttLine[3][iCount]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[3][iCount]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[3][(iCount - 1)] != NULL) {
									if(!uAttLine[3][(iCount - 1)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[3][(iCount - 1)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[3][(iCount + 1)] != NULL) {
									if(!uAttLine[3][(iCount + 1)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[3][(iCount + 1)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[3][(iCount - 2)] != NULL) {
									if(!uAttLine[3][(iCount - 2)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[3][(iCount - 2)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								} else if(uAttLine[3][(iCount + 2)] != NULL) {
									if(!uAttLine[3][(iCount + 2)]->isDead())
									{
										if(uDefLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[3][(iCount + 2)]->updateStackCombat(uDefLine[0][iCount], 2, false, bStack);
											bAttack = true;
										}
									}
								}
							}
						}
					}

					// Dale - Stack Attack: defender bombard
					for(iCount = 2; iCount < (iNumCols - 3); iCount++)
					{
						if(uDefLine[1][iCount] != NULL)
						{
							if(!uDefLine[1][iCount]->isDead())
							{
								if(uAttLine[1][iCount] != NULL)
								{
									if(!uAttLine[1][iCount]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[1][(iCount - 1)] != NULL) {
									if(!uAttLine[1][(iCount - 1)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[1][(iCount - 1)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[1][(iCount + 1)] != NULL) {
									if(!uAttLine[1][(iCount + 1)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[1][(iCount + 1)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[1][(iCount - 2)] != NULL) {
									if(!uAttLine[1][(iCount - 2)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[1][(iCount - 2)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[1][(iCount + 2)] != NULL) {
									if(!uAttLine[1][(iCount + 2)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[1][(iCount + 2)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[2][iCount] != NULL) {
									if(!uAttLine[2][iCount]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[2][(iCount - 1)] != NULL) {
									if(!uAttLine[2][(iCount - 1)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[2][(iCount - 1)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[2][(iCount + 1)] != NULL) {
									if(!uAttLine[2][(iCount + 1)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[2][(iCount + 1)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[2][(iCount - 2)] != NULL) {
									if(!uAttLine[2][(iCount - 2)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[2][(iCount - 2)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[2][(iCount + 2)] != NULL) {
									if(!uAttLine[2][(iCount + 2)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[2][(iCount + 2)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[3][iCount] != NULL) {
									if(!uAttLine[3][iCount]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
                                                uAttLine[3][iCount]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[3][(iCount - 1)] != NULL) {
									if(!uAttLine[3][(iCount - 1)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[3][(iCount - 1)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[3][(iCount + 1)] != NULL) {
									if(!uAttLine[3][(iCount + 1)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[3][(iCount + 1)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[3][(iCount - 2)] != NULL) {
									if(!uAttLine[3][(iCount - 2)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[3][(iCount - 2)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[3][(iCount + 2)] != NULL) {
									if(!uAttLine[3][(iCount + 2)]->isDead())
									{
										if(uDefLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] != NULL || uDefLine[2][iCount] != NULL)
											{
												uAttLine[3][(iCount + 2)]->updateStackCombat(uDefLine[1][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								}
							}
						}
					}

					// Dale - Stack Attack: defender ranged
					for(iCount = 2; iCount < (iNumCols - 3); iCount++)
					{
						if(uDefLine[2][iCount] != NULL)
						{
							if(!uDefLine[2][iCount]->isDead())
							{
								if(uAttLine[2][iCount] != NULL)
								{
									if(!uAttLine[2][iCount]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[2][(iCount - 1)] != NULL) {
									if(!uAttLine[2][(iCount - 1)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[2][(iCount - 1)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[2][(iCount + 1)] != NULL) {
									if(!uAttLine[2][(iCount + 1)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[2][(iCount + 1)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[2][(iCount - 2)] != NULL) {
									if(!uAttLine[2][(iCount - 2)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[2][(iCount - 2)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[2][(iCount + 2)] != NULL) {
									if(!uAttLine[2][(iCount + 2)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[2][(iCount + 2)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[1][iCount] != NULL) {
									if(!uAttLine[1][iCount]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[1][(iCount - 1)] != NULL) {
									if(!uAttLine[1][(iCount - 1)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[1][(iCount - 1)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[1][(iCount + 1)] != NULL) {
									if(!uAttLine[1][(iCount + 1)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[1][(iCount + 1)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[1][(iCount - 2)] != NULL) {
									if(!uAttLine[1][(iCount - 2)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[1][(iCount - 2)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[1][(iCount + 2)] != NULL) {
									if(!uAttLine[1][(iCount + 2)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[1][(iCount + 2)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[3][iCount] != NULL) {
									if(!uAttLine[3][iCount]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[3][iCount]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[3][(iCount - 1)] != NULL) {
									if(!uAttLine[3][(iCount - 1)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[3][(iCount - 1)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[3][(iCount + 1)] != NULL) {
									if(!uAttLine[3][(iCount + 1)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[3][(iCount + 1)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[3][(iCount - 2)] != NULL) {
									if(!uAttLine[3][(iCount - 2)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[3][(iCount - 2)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								} else if(uAttLine[3][(iCount + 2)] != NULL) {
									if(!uAttLine[3][(iCount + 2)]->isDead())
									{
										if(uDefLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] != NULL)
											{
												uAttLine[3][(iCount + 2)]->updateStackCombat(uDefLine[2][iCount], 1, false, bStack);
												bAttack = true;
											}
										}
									}
								}
							}
						}
					}

					// Dale - Stack Attack: attacker airstrike
					for(iCount = 2; iCount < (iNumCols - 3); iCount++)
					{
						if(uAttLine[0][iCount] != NULL)
						{
							if(!uAttLine[0][iCount]->isDead())
							{
								if(uDefLine[0][iCount] != NULL)
								{
									if(!uDefLine[0][iCount]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[0][iCount], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[0][(iCount - 1)] != NULL) {
									if(!uDefLine[0][(iCount - 1)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[0][(iCount - 1)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[0][(iCount + 1)] != NULL) {
									if(!uDefLine[0][(iCount + 1)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[0][(iCount + 1)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[0][(iCount - 2)] != NULL) {
									if(!uDefLine[0][(iCount - 2)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[0][(iCount - 2)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[0][(iCount + 2)] != NULL) {
									if(!uDefLine[0][(iCount + 2)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[0][(iCount + 2)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[1][iCount] != NULL) {
									if(!uDefLine[1][iCount]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[1][iCount], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount - 1)] != NULL) {
									if(!uDefLine[1][(iCount - 1)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[1][(iCount - 1)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount + 1)] != NULL) {
									if(!uDefLine[1][(iCount + 1)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[1][(iCount + 1)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount - 2)] != NULL) {
									if(!uDefLine[1][(iCount - 2)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[1][(iCount - 2)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount + 2)] != NULL) {
									if(!uDefLine[1][(iCount + 2)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[1][(iCount + 2)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[2][iCount] != NULL) {
									if(!uDefLine[2][iCount]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[2][iCount], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount - 1)] != NULL) {
									if(!uDefLine[2][(iCount - 1)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[2][(iCount - 1)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount + 1)] != NULL) {
									if(!uDefLine[2][(iCount + 1)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[2][(iCount + 1)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount - 2)] != NULL) {
									if(!uDefLine[2][(iCount - 2)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[2][(iCount - 2)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount + 2)] != NULL) {
									if(!uDefLine[2][(iCount + 2)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[2][(iCount + 2)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[3][iCount] != NULL) {
									if(!uDefLine[3][iCount]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[3][iCount], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount - 1)] != NULL) {
									if(!uDefLine[3][(iCount - 1)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[3][(iCount - 1)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount + 1)] != NULL) {
									if(!uDefLine[3][(iCount + 1)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[3][(iCount + 1)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount - 2)] != NULL) {
									if(!uDefLine[3][(iCount - 2)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[3][(iCount - 2)], 2, true, bStack);
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount + 2)] != NULL) {
									if(!uDefLine[3][(iCount + 2)]->isDead())
									{
										if(uAttLine[0][iCount]->getDomainType() == DOMAIN_AIR)
										{
											uAttLine[0][iCount]->updateStackCombat(uDefLine[3][(iCount + 2)], 2, true, bStack);
											bAttack = true;
										}
									}
								}
							}
						}
					}

					// Dale - Stack Attack: attacker bombard
					for(iCount = 2; iCount < (iNumCols - 3); iCount++)
					{
						if(uAttLine[1][iCount] != NULL)
						{
							if(!uAttLine[1][iCount]->isDead())
							{
								if(uDefLine[1][iCount] != NULL)
								{
									if(!uDefLine[1][iCount]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] == NULL && uDefLine[2][iCount] == NULL && uAttLine[3][iCount] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][iCount], 1, true, bStack);
											} else {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][iCount], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount - 1)] != NULL) {
									if(!uDefLine[1][(iCount - 1)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][(iCount - 1)] == NULL && uDefLine[2][(iCount - 1)] == NULL && uAttLine[3][(iCount - 1)] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][(iCount - 1)], 1, true, bStack);
											} else {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][(iCount - 1)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount + 1)] != NULL) {
									if(!uDefLine[1][(iCount + 1)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][(iCount + 1)] == NULL && uDefLine[2][(iCount + 1)] == NULL && uAttLine[3][(iCount + 1)] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][(iCount + 1)], 1, true, bStack);
											} else {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][(iCount + 1)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount - 2)] != NULL) {
									if(!uDefLine[1][(iCount - 2)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][(iCount - 2)] == NULL && uDefLine[2][(iCount - 2)] == NULL && uAttLine[3][(iCount - 2)] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][(iCount - 2)], 1, true, bStack);
											} else {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][(iCount - 2)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount + 2)] != NULL) {
									if(!uDefLine[1][(iCount + 2)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][(iCount + 2)] == NULL && uDefLine[2][(iCount + 2)] == NULL && uAttLine[3][(iCount + 2)] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][(iCount + 2)], 1, true, bStack);
											} else {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[1][(iCount + 2)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[2][iCount] != NULL) {
									if(!uDefLine[2][iCount]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][iCount] == NULL && uAttLine[3][iCount] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][iCount], 1, true, bStack);
											} else if(uAttLine[2][iCount] != NULL) {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][iCount], 1, true, bStack);
											} else {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][iCount], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount - 1)] != NULL) {
									if(!uDefLine[2][(iCount - 1)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][(iCount - 1)] == NULL && uAttLine[3][(iCount - 1)] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount - 1)], 1, true, bStack);
											} else if(uAttLine[2][(iCount - 1)] != NULL) {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount - 1)], 1, true, bStack);
											} else {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount - 1)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount + 1)] != NULL) {
									if(!uDefLine[2][(iCount + 1)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][(iCount + 1)] == NULL && uAttLine[3][(iCount + 1)] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount + 1)], 1, true, bStack);
											} else if(uAttLine[2][(iCount + 1)] != NULL) {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount + 1)], 1, true, bStack);
											} else {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount + 1)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount - 2)] != NULL) {
									if(!uDefLine[2][(iCount - 2)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][(iCount - 2)] == NULL && uAttLine[3][(iCount - 2)] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount - 2)], 1, true, bStack);
											} else if(uAttLine[2][(iCount - 2)] != NULL) {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount - 2)], 1, true, bStack);
											} else {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount - 2)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount + 2)] != NULL) {
									if(!uDefLine[2][(iCount + 2)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uDefLine[3][(iCount + 2)] == NULL && uAttLine[3][(iCount + 2)] != NULL)
											{
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount + 2)], 1, true, bStack);
											} else if(uAttLine[2][(iCount + 2)] != NULL) {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount + 2)], 1, true, bStack);
											} else {
												uAttLine[1][iCount]->updateStackCombat(uDefLine[2][(iCount + 2)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[3][iCount] != NULL) {
									if(!uDefLine[3][iCount]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uAttLine[3][iCount] == NULL)
											{
                                                uAttLine[1][iCount]->updateStackCombat(uDefLine[3][iCount], 2, true, !bStack);
											} else {
                                                uAttLine[1][iCount]->updateStackCombat(uDefLine[3][iCount], 1, true, bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount - 1)] != NULL) {
									if(!uDefLine[3][(iCount - 1)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uAttLine[3][iCount] == NULL)
											{
                                                uAttLine[1][iCount]->updateStackCombat(uDefLine[3][iCount], 2, true, !bStack);
											} else {
                                                uAttLine[1][iCount]->updateStackCombat(uDefLine[3][iCount], 1, true, bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount + 1)] != NULL) {
									if(!uDefLine[3][(iCount + 1)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uAttLine[3][(iCount + 1)] == NULL)
											{
                                                uAttLine[1][iCount]->updateStackCombat(uDefLine[3][(iCount + 1)], 2, true, !bStack);
											} else {
                                                uAttLine[1][iCount]->updateStackCombat(uDefLine[3][(iCount + 1)], 1, true, bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount - 2)] != NULL) {
									if(!uDefLine[3][(iCount - 2)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uAttLine[3][(iCount - 2)] == NULL)
											{
                                                uAttLine[1][iCount]->updateStackCombat(uDefLine[3][(iCount - 2)], 2, true, !bStack);
											} else {
                                                uAttLine[1][iCount]->updateStackCombat(uDefLine[3][(iCount - 2)], 1, true, bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount + 2)] != NULL) {
									if(!uDefLine[3][(iCount + 2)]->isDead())
									{
										if(uAttLine[1][iCount]->bombardRate() > 0)
										{
											if(uAttLine[3][(iCount + 2)] == NULL)
											{
                                                uAttLine[1][iCount]->updateStackCombat(uDefLine[3][(iCount + 2)], 2, true, !bStack);
											} else {
                                                uAttLine[1][iCount]->updateStackCombat(uDefLine[3][(iCount + 2)], 1, true, bStack);
											}
											bAttack = true;
										}
									}
								}
							}
						}
					}

					// Dale - Stack Attack: attacker ranged
					for(iCount = 2; iCount < (iNumCols - 3); iCount++)
					{
						if(uAttLine[2][iCount] != NULL)
						{
							if(!uAttLine[2][iCount]->isDead())
							{
								if(uDefLine[2][iCount] != NULL)
								{
									if(!uDefLine[2][iCount]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] == NULL && uAttLine[3][iCount] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][iCount], 1, true, bStack);
											} else {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][iCount], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount - 1)] != NULL) {
									if(!uDefLine[2][(iCount - 1)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][(iCount - 1)] == NULL && uAttLine[3][(iCount - 1)] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][(iCount - 1)], 1, true, bStack);
											} else {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][(iCount - 1)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount + 1)] != NULL) {
									if(!uDefLine[2][(iCount + 1)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][(iCount + 1)] == NULL && uAttLine[3][(iCount + 1)] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][(iCount + 1)], 1, true, bStack);
											} else {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][(iCount + 1)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount - 2)] != NULL) {
									if(!uDefLine[2][(iCount - 2)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][(iCount - 2)] == NULL && uAttLine[3][(iCount - 2)] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][(iCount - 2)], 1, true, bStack);
											} else {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][(iCount - 2)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[2][(iCount + 2)] != NULL) {
									if(!uDefLine[2][(iCount + 2)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][(iCount + 2)] == NULL && uAttLine[3][(iCount + 2)] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][(iCount + 2)], 1, true, bStack);
											} else {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[2][(iCount + 2)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[1][iCount] != NULL) {
									if(!uDefLine[1][iCount]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][iCount] == NULL && uDefLine[2][iCount] == NULL && uAttLine[3][iCount] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][iCount], 1, true, bStack);
											} else if(uAttLine[1][iCount] != NULL) {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][iCount], 1, true, bStack);
											} else {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][iCount], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount - 1)] != NULL) {
									if(!uDefLine[1][(iCount - 1)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][(iCount - 1)] == NULL && uDefLine[2][(iCount - 1)] == NULL && uAttLine[3][(iCount - 1)] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount - 1)], 1, true, bStack);
											} else if(uAttLine[1][(iCount - 1)] != NULL) {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount - 1)], 1, true, bStack);
											} else {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount - 1)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount + 1)] != NULL) {
									if(!uDefLine[1][(iCount + 1)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][(iCount + 1)] == NULL && uDefLine[2][(iCount + 1)] == NULL && uAttLine[3][(iCount + 1)] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount + 1)], 1, true, bStack);
											} else if(uAttLine[1][(iCount + 1)] != NULL) {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount + 1)], 1, true, bStack);
											} else {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount + 1)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount - 2)] != NULL) {
									if(!uDefLine[1][(iCount - 2)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][(iCount - 2)] == NULL && uDefLine[2][(iCount - 2)] == NULL && uAttLine[3][(iCount - 2)] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount - 2)], 1, true, bStack);
											} else if(uAttLine[1][(iCount - 2)] != NULL) {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount - 2)], 1, true, bStack);
											} else {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount - 2)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[1][(iCount + 2)] != NULL) {
									if(!uDefLine[1][(iCount + 2)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uDefLine[3][(iCount + 2)] == NULL && uDefLine[2][(iCount + 2)] == NULL && uAttLine[3][(iCount + 2)] != NULL)
											{
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount + 2)], 1, true, bStack);
											} else if(uAttLine[1][(iCount + 2)] != NULL) {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount + 2)], 1, true, bStack);
											} else {
												uAttLine[2][iCount]->updateStackCombat(uDefLine[1][(iCount + 2)], 2, true, !bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[3][iCount] != NULL) {
									if(!uDefLine[3][iCount]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uAttLine[3][iCount] == NULL)
											{
                                                uAttLine[2][iCount]->updateStackCombat(uDefLine[3][iCount], 2, true, !bStack);
											} else {
                                                uAttLine[2][iCount]->updateStackCombat(uDefLine[3][iCount], 1, true, bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount - 1)] != NULL) {
									if(!uDefLine[3][(iCount - 1)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uAttLine[3][(iCount - 1)] == NULL)
											{
                                                uAttLine[2][iCount]->updateStackCombat(uDefLine[3][(iCount - 1)], 2, true, !bStack);
											} else {
                                                uAttLine[2][iCount]->updateStackCombat(uDefLine[3][(iCount - 1)], 1, true, bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount + 1)] != NULL) {
									if(!uDefLine[3][(iCount + 1)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uAttLine[3][(iCount + 1)] == NULL)
											{
                                                uAttLine[2][iCount]->updateStackCombat(uDefLine[3][(iCount + 1)], 2, true, !bStack);
											} else {
                                                uAttLine[2][iCount]->updateStackCombat(uDefLine[3][(iCount + 1)], 1, true, bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount - 2)] != NULL) {
									if(!uDefLine[3][(iCount - 2)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uAttLine[3][(iCount - 2)] == NULL)
											{
                                                uAttLine[2][iCount]->updateStackCombat(uDefLine[3][(iCount - 2)], 2, true, !bStack);
											} else {
                                                uAttLine[2][iCount]->updateStackCombat(uDefLine[3][(iCount - 2)], 1, true, bStack);
											}
											bAttack = true;
										}
									}
								} else if(uDefLine[3][(iCount + 2)] != NULL) {
									if(!uDefLine[3][(iCount + 2)]->isDead())
									{
										if(uAttLine[2][iCount]->getDCMRanged())
										{
											if(uAttLine[3][(iCount + 2)] == NULL)
											{
                                                uAttLine[2][iCount]->updateStackCombat(uDefLine[3][(iCount + 2)], 2, true, !bStack);
											} else {
                                                uAttLine[2][iCount]->updateStackCombat(uDefLine[3][(iCount + 2)], 1, true, bStack);
											}
											bAttack = true;
										}
									}
								}
							}
						}
					}

					// Dale - stack Attack: attacker hand-to-hand combat
					for(iCount = 2; iCount < (iNumCols - 3); iCount++)
					{
						if(uAttLine[3][iCount] != NULL)
						{
							if(!uAttLine[3][iCount]->isDead())
							{
								if(uDefLine[3][iCount] != NULL)
								{
									if(!uDefLine[3][iCount]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[3][iCount], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[3][(iCount - 1)] != NULL) {
									if(!uDefLine[3][(iCount - 1)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[3][(iCount - 1)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[3][(iCount + 1)] != NULL) {
									if(!uDefLine[3][(iCount + 1)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[3][(iCount + 1)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[3][(iCount - 2)] != NULL) {
									if(!uDefLine[3][(iCount - 2)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[3][(iCount - 2)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[3][(iCount + 2)] != NULL) {
									if(!uDefLine[3][(iCount + 2)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[3][(iCount + 2)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[2][iCount] != NULL) {
									if(!uDefLine[2][iCount]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[2][iCount], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[2][(iCount - 1)] != NULL) {
									if(!uDefLine[2][(iCount - 1)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[2][(iCount - 1)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[2][(iCount + 1)] != NULL) {
									if(!uDefLine[2][(iCount + 1)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[2][(iCount + 1)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[2][(iCount - 2)] != NULL) {
									if(!uDefLine[2][(iCount - 2)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[2][(iCount - 2)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[2][(iCount + 2)] != NULL) {
									if(!uDefLine[2][(iCount + 2)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[2][(iCount + 2)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[1][iCount] != NULL) {
									if(!uDefLine[1][iCount]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[1][iCount], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[1][(iCount - 1)] != NULL) {
									if(!uDefLine[1][(iCount - 1)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[1][(iCount - 1)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[1][(iCount + 1)] != NULL) {
									if(!uDefLine[1][(iCount + 1)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[1][(iCount + 1)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[1][(iCount - 2)] != NULL) {
									if(!uDefLine[1][(iCount - 2)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[1][(iCount - 2)], 2, true, !bStack);
										bAttack = true;
									}
								} else if(uDefLine[1][(iCount + 2)] != NULL) {
									if(!uDefLine[1][(iCount + 2)]->isDead())
									{
										uAttLine[3][iCount]->updateStackCombat(uDefLine[1][(iCount + 2)], 2, true, !bStack);
										bAttack = true;
									}
								}
							}
						}
					}
				}
			}
		}
	}
	return bAttack;
}
